from __future__ import annotations

import asyncio
import collections
import json
import logging
import os
import socket
import time
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Coroutine, Optional, Sequence

if TYPE_CHECKING:
    from .progress_stream import TurnProgressTracker
    from .state import TelegramTopicRecord

from ...agents.opencode.supervisor import OpenCodeSupervisor
from ...core.flows.models import FlowRunRecord
from ...core.flows.pause_dispatch import format_pause_reason, latest_dispatch_seq
from ...core.hub import HubSupervisor
from ...core.locks import process_alive
from ...core.logging_utils import log_event
from ...core.request_context import reset_conversation_id, set_conversation_id
from ...core.runtime_services import RuntimeServices
from ...core.state import now_iso
from ...core.state_roots import resolve_global_state_root
from ...core.text_delta_coalescer import TextDeltaCoalescer
from ...core.utils import build_opencode_supervisor
from ...flows.ticket_flow.runtime_helpers import build_ticket_flow_runtime_resources
from ...housekeeping import HousekeepingConfig, run_housekeeping_for_roots
from ...integrations.app_server.threads import (
    AppServerThreadRegistry,
    default_app_server_threads_path,
)
from ...manifest import load_manifest
from ...tickets.replies import dispatch_reply, ensure_reply_dirs, resolve_reply_paths
from ...voice import VoiceConfig, VoiceService
from ..app_server.supervisor import WorkspaceAppServerSupervisor
from ..chat.service import ChatBotServiceCore
from ..chat.update_notifier import ChatUpdateStatusNotifier
from .adapter import (
    TelegramBotClient,
    TelegramCallbackQuery,
    TelegramDocument,
    TelegramMessage,
    TelegramPhotoSize,
    TelegramUpdate,
    TelegramUpdatePoller,
)
from .chat_adapter import TelegramChatAdapter
from .chat_state_store import TelegramChatStateStore
from .chat_transport import TelegramChatTransport
from .commands_registry import build_command_payloads, diff_command_lists
from .config import (
    AppServerUnavailableError,
    TelegramBotConfig,
    TelegramBotLockError,
    TelegramMediaCandidate,
)
from .config import TelegramBotConfigError as TelegramBotConfigError  # re-export
from .constants import (
    DEFAULT_INTERRUPT_TIMEOUT_SECONDS,
    QUEUED_PLACEHOLDER_TEXT,
    TurnKey,
)
from .dispatch import dispatch_update
from .handlers import callbacks as callback_handlers
from .handlers import messages as message_handlers
from .handlers.approvals import TelegramApprovalHandlers
from .handlers.commands import build_command_specs
from .handlers.commands_runtime import TelegramCommandHandlers
from .handlers.messages import _CoalescedBuffer
from .handlers.questions import TelegramQuestionHandlers
from .handlers.selections import TelegramSelectionHandlers
from .helpers import (
    ModelOption,
    _lock_payload_summary,
    _read_lock_payload,
    _split_topic_key,
    _telegram_lock_path,
    _with_conversation_id,
)
from .notifications import TelegramNotificationHandlers
from .outbox import TelegramOutboxManager
from .runtime import TelegramRuntimeHelpers
from .state import (
    TelegramStateStore,
    TopicRouter,
    parse_topic_key,
    topic_key,
)
from .ticket_flow_bridge import (
    TelegramTicketFlowBridge,
)
from .transport import TelegramMessageTransport
from .types import (
    CompactState,
    ModelPickerState,
    PendingApproval,
    PendingQuestion,
    ReviewCommitSelectionState,
    SelectionState,
    TurnContext,
)
from .voice import TelegramVoiceManager

TICKET_FLOW_WATCH_INTERVAL_SECONDS = 20


def _build_opencode_supervisor(
    config: TelegramBotConfig,
    *,
    logger: logging.Logger,
) -> Optional[OpenCodeSupervisor]:
    opencode_command = config.opencode_command or None
    opencode_binary = config.agent_binaries.get("opencode")

    supervisor = build_opencode_supervisor(
        opencode_command=opencode_command,
        opencode_binary=opencode_binary,
        workspace_root=config.root,
        logger=logger,
        request_timeout=None,
        max_handles=config.app_server_max_handles,
        idle_ttl_seconds=config.app_server_idle_ttl_seconds,
        base_env=None,
        subagent_models=None,
    )

    if supervisor is None:
        log_event(
            logger,
            logging.INFO,
            "telegram.opencode.unavailable",
            reason="command_missing",
        )
        return None

    return supervisor


def _next_reply_seq_sync(reply_history_dir: Any) -> int:
    from pathlib import Path

    path = Path(reply_history_dir)
    if not path.exists() or not path.is_dir():
        return 1
    existing: list[int] = []
    _SEQ_RE = __import__("re").compile(r"^[0-9]{4}$")
    for child in path.iterdir():
        try:
            if not child.is_dir():
                continue
            if not _SEQ_RE.fullmatch(child.name):
                continue
            existing.append(int(child.name))
        except OSError:
            continue
    return (max(existing) + 1) if existing else 1


class TelegramBotService(
    TelegramRuntimeHelpers,
    TelegramMessageTransport,
    TelegramNotificationHandlers,
    TelegramApprovalHandlers,
    TelegramQuestionHandlers,
    TelegramSelectionHandlers,
    TelegramCommandHandlers,
):
    TICKET_FLOW_WATCH_INTERVAL_SECONDS = TICKET_FLOW_WATCH_INTERVAL_SECONDS

    def __init__(
        self,
        config: TelegramBotConfig,
        *,
        logger: Optional[logging.Logger] = None,
        hub_root: Optional[Path] = None,
        manifest_path: Optional[Path] = None,
        voice_config: Optional[VoiceConfig] = None,
        voice_service: Optional[VoiceService] = None,
        housekeeping_config: Optional[HousekeepingConfig] = None,
        update_repo_url: Optional[str] = None,
        update_repo_ref: Optional[str] = None,
        update_skip_checks: bool = False,
        update_backend: str = "auto",
        update_linux_service_names: Optional[dict[str, str]] = None,
        app_server_auto_restart: Optional[bool] = None,
    ) -> None:
        self._config = config
        self._logger = logger or logging.getLogger(__name__)
        self._hub_root = hub_root
        self._manifest_path = manifest_path
        self._hub_supervisor = None
        self._hub_thread_registry = None
        if self._hub_root:
            try:
                self._hub_supervisor = HubSupervisor.from_path(self._hub_root)
            except Exception as exc:
                log_event(
                    self._logger,
                    logging.WARNING,
                    "telegram.pma.hub_supervisor.unavailable",
                    hub_root=str(self._hub_root),
                    exc=exc,
                )
            try:
                self._hub_thread_registry = AppServerThreadRegistry(
                    default_app_server_threads_path(self._hub_root)
                )
            except Exception as exc:
                log_event(
                    self._logger,
                    logging.WARNING,
                    "telegram.pma.thread_registry.unavailable",
                    hub_root=str(self._hub_root),
                    exc=exc,
                )
        self._update_repo_url = update_repo_url
        self._update_repo_ref = update_repo_ref
        self._update_skip_checks = update_skip_checks
        self._update_backend = update_backend
        self._update_linux_service_names = update_linux_service_names or {}
        self._app_server_auto_restart = app_server_auto_restart
        self._allowlist = config.allowlist()
        self._store = TelegramStateStore(
            config.state_file, default_approval_mode=config.defaults.approval_mode
        )
        self._router = TopicRouter(self._store)
        self._app_server_state_root = resolve_global_state_root() / "workspaces"
        self._app_server_supervisor = WorkspaceAppServerSupervisor(
            config.app_server_command,
            state_root=self._app_server_state_root,
            env_builder=self._build_workspace_env,
            approval_handler=self._handle_approval_request,
            notification_handler=self._handle_app_server_notification,
            logger=self._logger,
            auto_restart=self._app_server_auto_restart,
            max_handles=config.app_server_max_handles,
            idle_ttl_seconds=config.app_server_idle_ttl_seconds,
        )
        self._opencode_supervisor = _build_opencode_supervisor(
            config,
            logger=self._logger,
        )
        self._runtime_services = RuntimeServices(
            app_server_supervisor=self._app_server_supervisor,
            opencode_supervisor=self._opencode_supervisor,
            flow_runtime_builder=build_ticket_flow_runtime_resources,
        )
        poll_timeout = float(config.poll_timeout_seconds)
        request_timeout = config.poll_request_timeout_seconds
        if request_timeout is None:
            # Keep HTTP timeout above long-poll timeout to avoid ReadTimeout churn.
            request_timeout = max(poll_timeout + 5.0, 10.0)
        self._bot = TelegramBotClient(
            config.bot_token or "",
            logger=self._logger,
            timeout_seconds=float(request_timeout),
        )
        self._poller = TelegramUpdatePoller(
            self._bot, allowed_updates=config.poll_allowed_updates
        )
        self._chat_adapter = TelegramChatAdapter(
            self._bot,
            poller=self._poller,
        )
        self._chat_transport = TelegramChatTransport(self)
        self._chat_state_store = TelegramChatStateStore(self._store)
        self._chat_core = ChatBotServiceCore(
            owner=self,
            runtime_services=self._runtime_services,
            state_store=self._chat_state_store,
            adapter=self._chat_adapter,
            transport=self._chat_transport,
        )
        self._model_options: dict[str, ModelPickerState] = {}
        self._model_pending: dict[str, ModelOption] = {}
        self._model_catalog_cache: dict[str, tuple[Any, float]] = {}
        self._agent_options: dict[str, SelectionState] = {}
        self._voice_config = voice_config
        self._voice_service = voice_service
        self._housekeeping_config = housekeeping_config
        if self._voice_service is None and voice_config is not None:
            try:
                self._voice_service = VoiceService(voice_config, logger=self._logger)
            except Exception as exc:
                log_event(
                    self._logger,
                    logging.WARNING,
                    "telegram.voice.init_failed",
                    exc=exc,
                )
        self._turn_semaphore: Optional[asyncio.Semaphore] = None
        self._turn_contexts: dict[TurnKey, TurnContext] = {}
        self._reasoning_buffers: dict[str, TextDeltaCoalescer] = {}
        self._turn_preview_text: dict[TurnKey, str] = {}
        self._turn_preview_updated_at: dict[TurnKey, float] = {}
        self._turn_progress_trackers: dict[TurnKey, "TurnProgressTracker"] = {}
        self._turn_progress_rendered: dict[TurnKey, str] = {}
        self._turn_progress_updated_at: dict[TurnKey, float] = {}
        self._turn_progress_tasks: dict[TurnKey, asyncio.Task[None]] = {}
        self._turn_progress_heartbeat_tasks: dict[TurnKey, asyncio.Task[None]] = {}
        self._turn_progress_locks: dict[TurnKey, asyncio.Lock] = {}
        self._oversize_warnings: set[TurnKey] = set()
        self._pending_approvals: dict[str, PendingApproval] = {}
        self._pending_questions: dict[str, PendingQuestion] = {}
        self._ticket_flow_pause_targets: dict[str, str] = {}
        self._ticket_flow_bridge = TelegramTicketFlowBridge(
            logger=self._logger,
            store=self._store,
            pause_targets=self._ticket_flow_pause_targets,
            send_message_with_outbox=self._send_message_with_outbox,
            send_document=self._send_document,
            pause_config=self._config.pause_dispatch_notifications,
            default_notification_chat_id=self._config.default_notification_chat_id,
            hub_root=hub_root,
            manifest_path=manifest_path,
            config_root=self._config.root,
            runtime_services=self._runtime_services,
        )
        self._resume_options: dict[str, SelectionState] = {}
        self._bind_options: dict[str, SelectionState] = {}
        self._flow_run_options: dict[str, SelectionState] = {}
        self._update_options: dict[str, SelectionState] = {}
        self._update_confirm_options: dict[str, bool] = {}
        self._review_commit_options: dict[str, ReviewCommitSelectionState] = {}
        self._review_commit_subjects: dict[str, dict[str, str]] = {}
        self._pending_review_custom: dict[str, dict[str, Any]] = {}
        self._compact_pending: dict[str, CompactState] = {}
        self._coalesced_buffers: dict[str, _CoalescedBuffer] = {}
        self._coalesce_locks: dict[str, asyncio.Lock] = {}
        self._media_batch_buffers: dict[str, message_handlers._MediaBatchBuffer] = {}
        self._media_batch_locks: dict[str, asyncio.Lock] = {}
        self._outbox_inflight: set[str] = set()
        self._outbox_lock: Optional[asyncio.Lock] = None
        self._queued_placeholder_map: dict[tuple[int, int], int] = {}
        self._queued_placeholder_timestamps: dict[tuple[int, int], float] = {}
        self._bot_username: Optional[str] = None
        self._token_usage_by_thread: "collections.OrderedDict[str, dict[str, Any]]" = (
            collections.OrderedDict()
        )
        self._token_usage_by_turn: "collections.OrderedDict[str, dict[str, Any]]" = (
            collections.OrderedDict()
        )
        self._outbox_task: Optional[asyncio.Task[None]] = None
        self._cache_cleanup_task: Optional[asyncio.Task[None]] = None
        self._ticket_flow_watch_task: Optional[asyncio.Task[None]] = None
        self._cache_timestamps: dict[str, dict[object, float]] = {}
        self._last_update_ids: dict[str, int] = {}
        self._last_update_persisted_at: dict[str, float] = {}
        self._spawned_tasks: set[asyncio.Task[Any]] = set()
        self._update_status_notifier = ChatUpdateStatusNotifier(
            platform="telegram",
            logger=self._logger,
            read_status=self._read_update_status,
            send_notice=self._send_update_status_notice,
            spawn_task=self._spawn_task,
            mark_notified=self._mark_update_notified,
            format_status=self._format_update_status_message,
            running_message="Update still running. Use /update status for the latest state.",
        )
        self._outbox_manager = TelegramOutboxManager(
            self._store,
            send_message=self._send_message,
            edit_message_text=self._edit_message_text,
            delete_message=self._delete_message,
            logger=self._logger,
        )
        self._voice_manager = TelegramVoiceManager(
            self._config,
            self._store,
            voice_config=self._voice_config,
            voice_service=self._voice_service,
            send_message=self._send_message,
            edit_message_text=self._edit_message_text,
            send_progress_message=self._send_voice_progress_message,
            deliver_transcript=self._deliver_voice_transcript,
            download_file=self._download_telegram_file,
            logger=self._logger,
        )
        self._voice_task: Optional[asyncio.Task[None]] = None
        self._housekeeping_task: Optional[asyncio.Task[None]] = None
        self._command_specs = build_command_specs(self)
        self._instance_lock_path: Optional[Path] = None

    async def _housekeeping_roots(self) -> list[Path]:
        roots: set[Path] = set()
        try:
            state = await self._store.load()
            for record in state.topics.values():
                if isinstance(record.workspace_path, str) and record.workspace_path:
                    roots.add(Path(record.workspace_path).expanduser().resolve())
        except Exception as exc:
            log_event(
                self._logger,
                logging.WARNING,
                "telegram.housekeeping.state_failed",
                exc=exc,
            )
        if self._hub_root and self._manifest_path and self._manifest_path.exists():
            try:
                manifest = load_manifest(self._manifest_path, self._hub_root)
                for repo in manifest.repos:
                    roots.add((self._hub_root / repo.path).resolve())
            except Exception as exc:
                log_event(
                    self._logger,
                    logging.WARNING,
                    "telegram.housekeeping.manifest_failed",
                    exc=exc,
                )
        if self._config.root:
            roots.add(self._config.root.resolve())
        return sorted(roots)

    async def _gather_workspace_roots(self) -> list[Path]:
        roots: set[Path] = set()
        try:
            state = await self._store.load()
            for record in state.topics.values():
                if isinstance(record.workspace_path, str) and record.workspace_path:
                    roots.add(Path(record.workspace_path).expanduser().resolve())
        except Exception as exc:
            log_event(
                self._logger,
                logging.WARNING,
                "telegram.prewarm.state_failed",
                exc=exc,
            )
        return sorted(roots)

    async def _prewarm_workspace_clients(self) -> None:
        workspace_roots = await self._gather_workspace_roots()
        if not workspace_roots:
            log_event(
                self._logger,
                logging.INFO,
                "telegram.prewarm.skipped",
                reason="no_workspaces",
            )
            return

        log_event(
            self._logger,
            logging.INFO,
            "telegram.prewarm.started",
            workspace_count=len(workspace_roots),
            workspaces=[str(p) for p in workspace_roots],
        )

        sem = asyncio.Semaphore(3)
        prewarmed_count = 0
        failed_count = 0

        async def prewarm_one(workspace_root: Path) -> None:
            nonlocal prewarmed_count, failed_count
            async with sem:
                try:
                    await self._app_server_supervisor.get_client(workspace_root)
                    prewarmed_count += 1
                    log_event(
                        self._logger,
                        logging.INFO,
                        "telegram.prewarm.client_ready",
                        workspace_root=str(workspace_root),
                    )
                except Exception as exc:
                    failed_count += 1
                    log_event(
                        self._logger,
                        logging.WARNING,
                        "telegram.prewarm.client_failed",
                        workspace_root=str(workspace_root),
                        exc=exc,
                    )

        await asyncio.gather(
            *[prewarm_one(root) for root in workspace_roots],
            return_exceptions=True,
        )

        log_event(
            self._logger,
            logging.INFO,
            "telegram.prewarm.completed",
            workspace_count=len(workspace_roots),
            prewarmed_count=prewarmed_count,
            failed_count=failed_count,
        )

    async def _housekeeping_loop(self) -> None:
        config = self._housekeeping_config
        if config is None or not config.enabled:
            return
        interval = max(config.interval_seconds, 1)
        while True:
            try:
                roots = await self._housekeeping_roots()
                if roots:
                    await asyncio.to_thread(
                        run_housekeeping_for_roots,
                        config,
                        roots,
                        self._logger,
                    )
                await self._app_server_supervisor.prune_idle()
                if self._opencode_supervisor is not None:
                    await self._opencode_supervisor.prune_idle()
            except Exception as exc:
                log_event(
                    self._logger,
                    logging.WARNING,
                    "telegram.housekeeping.failed",
                    exc=exc,
                )
            await asyncio.sleep(interval)

    def _ensure_outbox_lock(self) -> asyncio.Lock:
        loop = asyncio.get_running_loop()
        lock = self._outbox_lock
        lock_loop = getattr(lock, "_loop", None) if lock else None
        if (
            lock is None
            or lock_loop is None
            or lock_loop is not loop
            or lock_loop.is_closed()
        ):
            lock = asyncio.Lock()
            self._outbox_lock = lock
        return lock

    async def _mark_outbox_inflight(self, record_id: str) -> bool:
        lock = self._ensure_outbox_lock()
        async with lock:
            if record_id in self._outbox_inflight:
                return False
            self._outbox_inflight.add(record_id)
            return True

    async def _clear_outbox_inflight(self, record_id: str) -> None:
        lock = self._ensure_outbox_lock()
        async with lock:
            self._outbox_inflight.discard(record_id)

    def _acquire_instance_lock(self) -> None:
        token = self._config.bot_token
        if not token:
            raise TelegramBotLockError("missing telegram bot token")
        lock_path = _telegram_lock_path(token)
        payload = {
            "pid": os.getpid(),
            "started_at": now_iso(),
            "host": socket.gethostname(),
            "cwd": os.getcwd(),
            "config_root": str(self._config.root),
        }
        lock_path.parent.mkdir(parents=True, exist_ok=True)
        try:
            fd = os.open(lock_path, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
        except FileExistsError as exc:
            existing = _read_lock_payload(lock_path)
            pid = existing.get("pid") if isinstance(existing, dict) else None
            if isinstance(pid, int) and process_alive(pid):
                log_event(
                    self._logger,
                    logging.ERROR,
                    "telegram.lock.contended",
                    lock_path=str(lock_path),
                    **_lock_payload_summary(existing),
                )
                raise TelegramBotLockError(
                    "Telegram bot already running for this token."
                ) from exc
            try:
                lock_path.unlink()
            except OSError:
                pass
            try:
                fd = os.open(lock_path, os.O_CREAT | os.O_EXCL | os.O_WRONLY)
            except FileExistsError as exc:
                existing = _read_lock_payload(lock_path)
                log_event(
                    self._logger,
                    logging.ERROR,
                    "telegram.lock.contended",
                    lock_path=str(lock_path),
                    **_lock_payload_summary(existing),
                )
                raise TelegramBotLockError(
                    "Telegram bot already running for this token."
                ) from exc
        with os.fdopen(fd, "w", encoding="utf-8") as handle:
            handle.write(json.dumps(payload) + "\n")
        self._instance_lock_path = lock_path
        log_event(
            self._logger,
            logging.INFO,
            "telegram.lock.acquired",
            lock_path=str(lock_path),
            **_lock_payload_summary(payload),
        )

    def _release_instance_lock(self) -> None:
        lock_path = self._instance_lock_path
        if lock_path is None:
            return
        existing = _read_lock_payload(lock_path)
        if isinstance(existing, dict):
            pid = existing.get("pid")
            if isinstance(pid, int) and pid != os.getpid():
                return
        try:
            lock_path.unlink()
        except OSError:
            pass
        self._instance_lock_path = None

    def _ensure_turn_semaphore(self) -> asyncio.Semaphore:
        if self._turn_semaphore is None:
            self._turn_semaphore = asyncio.Semaphore(
                self._config.concurrency.max_parallel_turns
            )
        return self._turn_semaphore

    async def run_polling(self) -> None:
        await self._chat_core.run()

    async def _dispatch_update(self, update: TelegramUpdate) -> None:
        await dispatch_update(self, update)

    async def _prime_bot_identity(self) -> None:
        try:
            payload = await self._bot.get_me()
        except Exception:
            return
        if isinstance(payload, dict):
            username = payload.get("username")
            if isinstance(username, str) and username:
                self._bot_username = username

    async def _register_bot_commands(self) -> None:
        registration = self._config.command_registration
        if not registration.enabled:
            log_event(
                self._logger,
                logging.DEBUG,
                "telegram.commands.disabled",
            )
            return
        desired, invalid = build_command_payloads(self._command_specs)
        if invalid:
            log_event(
                self._logger,
                logging.WARNING,
                "telegram.commands.invalid",
                invalid=invalid,
            )
        if not desired:
            log_event(
                self._logger,
                logging.WARNING,
                "telegram.commands.empty",
            )
            return
        if len(desired) > 100:
            log_event(
                self._logger,
                logging.WARNING,
                "telegram.commands.truncated",
                desired_count=len(desired),
            )
            desired = desired[:100]
        for scope_spec in registration.scopes:
            scope = scope_spec.scope
            language_code = scope_spec.language_code
            try:
                current = await self._bot.get_my_commands(
                    scope=scope,
                    language_code=language_code,
                )
            except Exception as exc:
                log_event(
                    self._logger,
                    logging.WARNING,
                    "telegram.commands.get_failed",
                    scope=scope,
                    language_code=language_code,
                    exc=exc,
                )
                continue
            diff = diff_command_lists(desired, current)
            if not diff.needs_update:
                log_event(
                    self._logger,
                    logging.DEBUG,
                    "telegram.commands.up_to_date",
                    scope=scope,
                    language_code=language_code,
                )
                continue
            try:
                updated = await self._bot.set_my_commands(
                    desired,
                    scope=scope,
                    language_code=language_code,
                )
            except Exception as exc:
                log_event(
                    self._logger,
                    logging.WARNING,
                    "telegram.commands.set_failed",
                    scope=scope,
                    language_code=language_code,
                    exc=exc,
                )
                continue
            log_event(
                self._logger,
                logging.INFO,
                "telegram.commands.updated",
                scope=scope,
                language_code=language_code,
                updated=updated,
                added=diff.added,
                removed=diff.removed,
                changed=diff.changed,
                order_changed=diff.order_changed,
            )

    async def _prime_poller_offset(self) -> None:
        last_update_id = await self._store.get_last_update_id_global()
        if not isinstance(last_update_id, int) or isinstance(last_update_id, bool):
            return
        offset = last_update_id + 1
        self._poller.set_offset(offset)
        log_event(
            self._logger,
            logging.INFO,
            "telegram.poll.offset.init",
            stored_global_update_id=last_update_id,
            poller_offset=offset,
        )

    async def _record_poll_offset(self, updates: Sequence[TelegramUpdate]) -> None:
        offset = self._poller.offset
        if offset is None:
            return
        last_update_id = offset - 1
        if last_update_id < 0:
            return
        stored = await self._store.update_last_update_id_global(last_update_id)
        max_update_id = max((update.update_id for update in updates), default=None)
        log_event(
            self._logger,
            logging.INFO,
            "telegram.poll.offset.updated",
            incoming_update_id=max_update_id,
            stored_global_update_id=stored,
            poller_offset=offset,
        )

    def _spawn_task(self, coro: Coroutine[Any, Any, Any]) -> asyncio.Task[Any]:
        task: asyncio.Task[Any] = asyncio.create_task(coro)
        self._spawned_tasks.add(task)
        task.add_done_callback(self._log_task_result)
        return task

    def _log_task_result(self, task: asyncio.Future) -> None:
        if isinstance(task, asyncio.Task):
            self._spawned_tasks.discard(task)
        try:
            task.result()
        except asyncio.CancelledError:
            return
        except Exception as exc:
            log_event(self._logger, logging.WARNING, "telegram.task.failed", exc=exc)

    def _touch_cache_timestamp(self, cache_name: str, key: object) -> None:
        cache = self._cache_timestamps.setdefault(cache_name, {})
        cache[key] = time.monotonic()

    def _evict_expired_cache_entries(self, cache_name: str, ttl_seconds: float) -> None:
        cache = self._cache_timestamps.get(cache_name)
        if not cache:
            return
        now = time.monotonic()
        expired: list[object] = []
        for key, updated_at in cache.items():
            if (now - updated_at) > ttl_seconds:
                expired.append(key)
        if not expired:
            return
        for key in expired:
            cache.pop(key, None)
            if cache_name == "reasoning_buffers":
                self._reasoning_buffers.pop(key, None)
            elif cache_name == "turn_preview":
                self._turn_preview_text.pop(key, None)
                self._turn_preview_updated_at.pop(key, None)
            elif cache_name == "progress_trackers":
                self._turn_progress_trackers.pop(key, None)
                self._turn_progress_rendered.pop(key, None)
                self._turn_progress_updated_at.pop(key, None)
                task = self._turn_progress_tasks.pop(key, None)
                if task and not task.done():
                    task.cancel()
                heartbeat_task = self._turn_progress_heartbeat_tasks.pop(key, None)
                if heartbeat_task and not heartbeat_task.done():
                    heartbeat_task.cancel()
            elif cache_name == "oversize_warnings":
                self._oversize_warnings.discard(key)
            elif cache_name == "coalesced_buffers":
                self._coalesced_buffers.pop(key, None)
                self._coalesce_locks.pop(key, None)
            elif cache_name == "media_batch_buffers":
                self._media_batch_buffers.pop(key, None)
                self._media_batch_locks.pop(key, None)
            elif cache_name == "resume_options":
                self._resume_options.pop(key, None)
            elif cache_name == "bind_options":
                self._bind_options.pop(key, None)
            elif cache_name == "flow_run_options":
                self._flow_run_options.pop(key, None)
            elif cache_name == "agent_options":
                self._agent_options.pop(key, None)
            elif cache_name == "update_options":
                self._update_options.pop(key, None)
            elif cache_name == "update_confirm_options":
                self._update_confirm_options.pop(key, None)
            elif cache_name == "review_commit_options":
                self._review_commit_options.pop(key, None)
            elif cache_name == "review_commit_subjects":
                self._review_commit_subjects.pop(key, None)
            elif cache_name == "pending_review_custom":
                self._pending_review_custom.pop(key, None)
            elif cache_name == "compact_pending":
                self._compact_pending.pop(key, None)
            elif cache_name == "model_options":
                self._model_options.pop(key, None)
            elif cache_name == "model_pending":
                self._model_pending.pop(key, None)
            elif cache_name == "pending_approvals":
                self._pending_approvals.pop(key, None)
            elif cache_name == "pending_questions":
                self._pending_questions.pop(key, None)

    async def _cache_cleanup_loop(self) -> None:
        interval = max(self._config.cache.cleanup_interval_seconds, 1.0)
        while True:
            await asyncio.sleep(interval)
            self._evict_expired_cache_entries(
                "reasoning_buffers", self._config.cache.reasoning_buffer_ttl_seconds
            )
            self._evict_expired_cache_entries(
                "turn_preview", self._config.cache.turn_preview_ttl_seconds
            )
            self._evict_expired_cache_entries(
                "progress_trackers", self._config.cache.progress_stream_ttl_seconds
            )
            self._evict_expired_cache_entries(
                "oversize_warnings", self._config.cache.oversize_warning_ttl_seconds
            )
            self._evict_expired_cache_entries(
                "coalesced_buffers", self._config.cache.coalesce_buffer_ttl_seconds
            )
            self._evict_expired_cache_entries(
                "media_batch_buffers",
                self._config.cache.media_batch_buffer_ttl_seconds,
            )
            self._evict_expired_cache_entries(
                "resume_options", self._config.cache.selection_state_ttl_seconds
            )
            self._evict_expired_cache_entries(
                "bind_options", self._config.cache.selection_state_ttl_seconds
            )
            self._evict_expired_cache_entries(
                "flow_run_options", self._config.cache.selection_state_ttl_seconds
            )
            self._evict_expired_cache_entries(
                "agent_options", self._config.cache.selection_state_ttl_seconds
            )
            self._evict_expired_cache_entries(
                "update_options", self._config.cache.selection_state_ttl_seconds
            )
            self._evict_expired_cache_entries(
                "update_confirm_options",
                self._config.cache.selection_state_ttl_seconds,
            )
            self._evict_expired_cache_entries(
                "review_commit_options",
                self._config.cache.selection_state_ttl_seconds,
            )
            self._evict_expired_cache_entries(
                "review_commit_subjects",
                self._config.cache.selection_state_ttl_seconds,
            )
            self._evict_expired_cache_entries(
                "pending_review_custom",
                self._config.cache.selection_state_ttl_seconds,
            )
            self._evict_expired_cache_entries(
                "compact_pending", self._config.cache.selection_state_ttl_seconds
            )
            self._evict_expired_cache_entries(
                "model_options", self._config.cache.selection_state_ttl_seconds
            )
            self._evict_expired_cache_entries(
                "model_pending", self._config.cache.model_pending_ttl_seconds
            )
            self._evict_expired_cache_entries(
                "pending_approvals", self._config.cache.pending_approval_ttl_seconds
            )
            self._evict_expired_cache_entries(
                "pending_questions", self._config.cache.pending_question_ttl_seconds
            )
            now = time.monotonic()
            expired_placeholders = []
            for key, timestamp in self._queued_placeholder_timestamps.items():
                if (now - timestamp) > self._config.cache.pending_approval_ttl_seconds:
                    expired_placeholders.append(key)
            for key in expired_placeholders:
                self._queued_placeholder_map.pop(key, None)
                self._queued_placeholder_timestamps.pop(key, None)

    @staticmethod
    def _parse_last_active(record: "TelegramTopicRecord") -> float:
        raw = getattr(record, "last_active_at", None)
        if isinstance(raw, str):
            try:
                return datetime.strptime(raw, "%Y-%m-%dT%H:%M:%SZ").timestamp()
            except ValueError:
                return float("-inf")
        return float("-inf")

    def _select_ticket_flow_topic(
        self, entries: list[tuple[str, "TelegramTopicRecord"]]
    ) -> Optional[tuple[str, "TelegramTopicRecord"]]:
        return self._ticket_flow_bridge._select_ticket_flow_topic(entries)

    @staticmethod
    def _set_ticket_dispatch_marker(
        value: Optional[str],
    ) -> "Callable[[TelegramTopicRecord], None]":
        def apply(topic: "TelegramTopicRecord") -> None:
            topic.last_ticket_dispatch_seq = value

        return apply

    async def _ticket_flow_watch_loop(self) -> None:
        await self._ticket_flow_bridge.watch_ticket_flow_pauses(
            TICKET_FLOW_WATCH_INTERVAL_SECONDS
        )

    async def _watch_ticket_flow_pauses(self) -> None:
        await self._ticket_flow_bridge._scan_and_notify_pauses()

    async def _notify_ticket_flow_pause(
        self,
        workspace_root: Path,
        entries: list[tuple[str, "TelegramTopicRecord"]],
    ) -> None:
        await self._ticket_flow_bridge._notify_ticket_flow_pause(
            workspace_root, entries
        )

    def _load_ticket_flow_pause(
        self, workspace_root: Path
    ) -> Optional[tuple[str, str, str, Optional[Path]]]:
        return self._ticket_flow_bridge._load_ticket_flow_pause(workspace_root)

    def _latest_dispatch_seq(self, history_dir: Path) -> Optional[str]:
        return latest_dispatch_seq(history_dir)

    def _format_ticket_flow_pause_reason(self, record: "FlowRunRecord") -> str:
        return format_pause_reason(record)

    def _get_paused_ticket_flow(
        self, workspace_root: Path, preferred_run_id: Optional[str] = None
    ) -> Optional[tuple[str, FlowRunRecord]]:
        return self._ticket_flow_bridge.get_paused_ticket_flow(
            workspace_root, preferred_run_id=preferred_run_id
        )

    async def _write_user_reply_from_telegram(
        self,
        workspace_root: Path,
        run_id: str,
        run_record: FlowRunRecord,
        message: TelegramMessage,
        text: str,
        files: Optional[list[tuple[str, bytes]]] = None,
    ) -> tuple[bool, str]:
        try:
            input_data = dict(run_record.input_data or {})
            runs_dir_raw = input_data.get("runs_dir")
            runs_dir = (
                Path(runs_dir_raw)
                if isinstance(runs_dir_raw, str) and runs_dir_raw
                else Path(".codex-autorunner/runs")
            )
            reply_paths = resolve_reply_paths(
                workspace_root=workspace_root, runs_dir=runs_dir, run_id=run_id
            )
            ensure_reply_dirs(reply_paths)

            cleaned_text = text.strip()
            raw = cleaned_text
            if raw and not raw.endswith("\n"):
                raw += "\n"

            await asyncio.to_thread(
                reply_paths.user_reply_path.write_text, raw, encoding="utf-8"
            )

            if files:
                for filename, data in files:
                    safe_name = Path(filename).name
                    if not safe_name or safe_name in {".", ".."}:
                        safe_name = "attachment"
                    dest = reply_paths.reply_dir / safe_name
                    dest.parent.mkdir(parents=True, exist_ok=True)
                    await asyncio.to_thread(dest.write_bytes, data)

            seq = await asyncio.to_thread(
                lambda: _next_reply_seq_sync(reply_paths.reply_history_dir)
            )
            dispatch, errors = await asyncio.to_thread(
                dispatch_reply, reply_paths, next_seq=seq
            )
            if errors:
                return False, "\n".join(errors)
            if dispatch is None:
                return False, "Failed to archive reply"
            return (
                True,
                f"Reply archived (seq {dispatch.seq}). Use /flow resume to continue.",
            )
        except Exception as exc:
            self._logger.warning(
                "Failed to write USER_REPLY.md from Telegram",
                exc=exc,
                workspace_root=str(workspace_root),
                run_id=run_id,
            )
            return False, f"Failed to write reply: {exc}"

    async def _interrupt_timeout_check(
        self, key: str, turn_id: str, message_id: int
    ) -> None:
        await asyncio.sleep(DEFAULT_INTERRUPT_TIMEOUT_SECONDS)
        runtime = self._router.runtime_for(key)
        if runtime.current_turn_id != turn_id:
            return
        if runtime.interrupt_message_id != message_id:
            return
        if runtime.interrupt_turn_id != turn_id:
            return
        chat_id, _thread_id = _split_topic_key(key)
        await self._edit_message_text(
            chat_id,
            message_id,
            "Still stopping... (30s). If this is stuck, try /interrupt again.",
        )
        runtime.interrupt_requested = False

    async def _dispatch_interrupt_request(
        self,
        *,
        turn_id: str,
        codex_thread_id: Optional[str],
        runtime: Any,
        chat_id: int,
        thread_id: Optional[int],
    ) -> None:
        key = await self._resolve_topic_key(chat_id, thread_id)
        record = await self._router.get_topic(key)
        turn_ctx = self._resolve_turn_context(turn_id, thread_id=codex_thread_id)
        if turn_ctx is not None and turn_ctx.topic_key and turn_ctx.topic_key != key:
            scoped_record = await self._router.get_topic(turn_ctx.topic_key)
            if scoped_record is not None:
                key = turn_ctx.topic_key
                record = scoped_record
        workspace_path = record.workspace_path if record else None
        if record and getattr(record, "pma_enabled", False):
            # PMA topics are hub-scoped; their stored topic records may not carry a
            # workspace path. Turn execution patches workspace_path to hub_root, but
            # interrupt dispatch must do the same so /interrupt works in PMA mode.
            hub_root = getattr(self, "_hub_root", None)
            if hub_root is not None:
                workspace_path = str(hub_root)
        if record and record.agent == "opencode":
            session_id = record.active_thread_id or codex_thread_id
            if (
                not session_id
                or self._opencode_supervisor is None
                or not isinstance(workspace_path, str)
                or not workspace_path
            ):
                runtime.interrupt_requested = False
                if runtime.interrupt_message_id is not None:
                    await self._edit_message_text(
                        chat_id,
                        runtime.interrupt_message_id,
                        "Interrupt failed (OpenCode unavailable).",
                    )
                    runtime.interrupt_message_id = None
                    runtime.interrupt_turn_id = None
                return
            try:
                client = await self._opencode_supervisor.get_client(
                    Path(workspace_path)
                )
                await client.abort(session_id)
            except Exception as exc:
                log_event(
                    self._logger,
                    logging.WARNING,
                    "telegram.interrupt.failed",
                    chat_id=chat_id,
                    thread_id=thread_id,
                    turn_id=turn_id,
                    exc=exc,
                )
                if (
                    runtime.interrupt_message_id is not None
                    and runtime.interrupt_turn_id == turn_id
                ):
                    await self._edit_message_text(
                        chat_id,
                        runtime.interrupt_message_id,
                        "Interrupt failed (OpenCode error).",
                    )
                runtime.interrupt_message_id = None
                runtime.interrupt_turn_id = None
                runtime.interrupt_requested = False
                return
            runtime.interrupt_message_id = None
            runtime.interrupt_turn_id = None
            runtime.interrupt_requested = False
            return
        try:
            client = await self._client_for_workspace(workspace_path)
        except AppServerUnavailableError:
            runtime.interrupt_requested = False
            if runtime.interrupt_message_id is not None:
                await self._edit_message_text(
                    chat_id,
                    runtime.interrupt_message_id,
                    "Interrupt failed (app-server unavailable).",
                )
                runtime.interrupt_message_id = None
                runtime.interrupt_turn_id = None
            return
        if client is None:
            runtime.interrupt_requested = False
            if runtime.interrupt_message_id is not None:
                await self._edit_message_text(
                    chat_id,
                    runtime.interrupt_message_id,
                    "Interrupt failed (app-server error).",
                )
                runtime.interrupt_message_id = None
                runtime.interrupt_turn_id = None
            return
        try:
            await client.turn_interrupt(turn_id, thread_id=codex_thread_id)
        except Exception as exc:
            log_event(
                self._logger,
                logging.WARNING,
                "telegram.interrupt.failed",
                chat_id=chat_id,
                thread_id=thread_id,
                turn_id=turn_id,
                exc=exc,
            )
            if (
                runtime.interrupt_message_id is not None
                and runtime.interrupt_turn_id == turn_id
            ):
                await self._edit_message_text(
                    chat_id,
                    runtime.interrupt_message_id,
                    "Interrupt failed (app-server error).",
                )
                runtime.interrupt_message_id = None
                runtime.interrupt_turn_id = None
            runtime.interrupt_requested = False

    async def _handle_message(self, message: TelegramMessage) -> None:
        await message_handlers.handle_message(self, message)

    def _should_bypass_topic_queue(self, message: TelegramMessage) -> bool:
        return message_handlers.should_bypass_topic_queue(self, message)

    async def _handle_edited_message(self, message: TelegramMessage) -> None:
        await message_handlers.handle_edited_message(self, message)

    async def _handle_message_inner(
        self,
        message: TelegramMessage,
        *,
        topic_key: Optional[str] = None,
        placeholder_id: Optional[int] = None,
    ) -> None:
        await message_handlers.handle_message_inner(
            self, message, topic_key=topic_key, placeholder_id=placeholder_id
        )

    def _coalesce_key_for_topic(self, key: str, user_id: Optional[int]) -> str:
        return message_handlers.coalesce_key_for_topic(self, key, user_id)

    async def _coalesce_key(self, message: TelegramMessage) -> str:
        return await message_handlers.coalesce_key(self, message)

    async def _buffer_coalesced_message(
        self,
        message: TelegramMessage,
        text: str,
        *,
        placeholder_id: Optional[int] = None,
    ) -> None:
        await message_handlers.buffer_coalesced_message(
            self, message, text, placeholder_id=placeholder_id
        )

    async def _coalesce_flush_after(self, key: str) -> None:
        window_seconds = self._config.coalesce_window_seconds
        await message_handlers.coalesce_flush_after(self, key, window_seconds)

    async def _flush_coalesced_message(self, message: TelegramMessage) -> None:
        await message_handlers.flush_coalesced_message(self, message)

    async def _flush_coalesced_key(self, key: str) -> None:
        await message_handlers.flush_coalesced_key(self, key)

    def _build_coalesced_message(self, buffer: _CoalescedBuffer) -> TelegramMessage:
        return message_handlers.build_coalesced_message(buffer)

    def _message_has_media(self, message: TelegramMessage) -> bool:
        return message_handlers.message_has_media(message)

    def _select_photo(
        self, photos: Sequence[TelegramPhotoSize]
    ) -> Optional[TelegramPhotoSize]:
        return message_handlers.select_photo(photos)

    def _document_is_image(self, document: TelegramDocument) -> bool:
        return message_handlers.document_is_image(document)

    def _select_image_candidate(
        self, message: TelegramMessage
    ) -> Optional[TelegramMediaCandidate]:
        return message_handlers.select_image_candidate(message)

    def _select_voice_candidate(
        self, message: TelegramMessage
    ) -> Optional[TelegramMediaCandidate]:
        return message_handlers.select_voice_candidate(message)

    async def _handle_media_message(
        self,
        message: TelegramMessage,
        runtime: Any,
        caption_text: str,
        *,
        placeholder_id: Optional[int] = None,
    ) -> None:
        await message_handlers.handle_media_message(
            self,
            message,
            runtime,
            caption_text,
            placeholder_id=placeholder_id,
        )

    def _with_conversation_id(
        self, message: str, *, chat_id: int, thread_id: Optional[int]
    ) -> str:
        return _with_conversation_id(message, chat_id=chat_id, thread_id=thread_id)

    async def _should_process_update(self, key: str, update_id: int) -> bool:
        if not isinstance(update_id, int):
            return True
        if isinstance(update_id, bool):
            return True
        last_id = self._last_update_ids.get(key)
        if last_id is None:
            record = None
            try:
                record = await self._store.get_topic(key)
            except Exception as exc:
                log_event(
                    self._logger,
                    logging.WARNING,
                    "telegram.update_id.load.failed",
                    exc=exc,
                    topic_key=key,
                    update_id=update_id,
                )
            last_id = record.last_update_id if record else None
            if isinstance(last_id, int) and not isinstance(last_id, bool):
                self._last_update_ids[key] = last_id
            else:
                last_id = None
        if isinstance(last_id, int) and update_id <= last_id:
            return False
        self._last_update_ids[key] = update_id
        try:
            await self._maybe_persist_update_id(key, update_id)
        except Exception as exc:
            log_event(
                self._logger,
                logging.WARNING,
                "telegram.update_id.persist.failed",
                exc=exc,
                topic_key=key,
                update_id=update_id,
            )
        return True

    async def _maybe_persist_update_id(self, key: str, update_id: int) -> None:
        now = time.monotonic()
        last_persisted = self._last_update_persisted_at.get(key, 0.0)
        if (
            now - last_persisted
        ) < self._config.cache.update_id_persist_interval_seconds:
            return

        def apply(record: "TelegramTopicRecord") -> None:
            record.last_update_id = update_id

        await self._store.update_topic(key, apply)
        self._last_update_persisted_at[key] = now

    async def _handle_callback(self, callback: TelegramCallbackQuery) -> None:
        await callback_handlers.handle_callback(self, callback)

    def _enqueue_topic_work(
        self, key: str, work: Any, *, force_queue: bool = False
    ) -> None:
        runtime = self._router.runtime_for(key)
        wrapped = self._wrap_topic_work(key, work)
        if force_queue or self._config.concurrency.per_topic_queue:
            self._spawn_task(runtime.queue.enqueue(wrapped))
        else:
            self._spawn_task(wrapped())

    async def _maybe_send_queued_placeholder(
        self, message: TelegramMessage, *, topic_key: str
    ) -> Optional[int]:
        runtime = self._router.runtime_for(topic_key)
        is_busy = runtime.current_turn_id is not None or runtime.queue.pending() > 0
        if not is_busy:
            return None
        placeholder_id = await self._send_placeholder(
            message.chat_id,
            thread_id=message.thread_id,
            reply_to=message.message_id,
            text=QUEUED_PLACEHOLDER_TEXT,
        )
        if placeholder_id is None:
            return None
        self._set_queued_placeholder(
            message.chat_id, message.message_id, placeholder_id
        )
        log_event(
            self._logger,
            logging.INFO,
            "telegram.placeholder.queued",
            topic_key=topic_key,
            chat_id=message.chat_id,
            thread_id=message.thread_id,
            message_id=message.message_id,
            placeholder_id=placeholder_id,
        )
        return placeholder_id

    def _wrap_placeholder_work(
        self,
        *,
        chat_id: int,
        placeholder_id: Optional[int],
        work: Any,
    ) -> Any:
        if placeholder_id is None:
            return work

        async def wrapped() -> Any:
            try:
                return await work()
            finally:
                await self._delete_message(chat_id, placeholder_id)

        return wrapped

    def _claim_queued_placeholder(self, chat_id: int, message_id: int) -> Optional[int]:
        placeholder_id = self._queued_placeholder_map.pop((chat_id, message_id), None)
        self._queued_placeholder_timestamps.pop((chat_id, message_id), None)
        return placeholder_id

    def _get_queued_placeholder(self, chat_id: int, message_id: int) -> Optional[int]:
        return self._queued_placeholder_map.get((chat_id, message_id))

    def _set_queued_placeholder(
        self, chat_id: int, message_id: int, placeholder_id: int
    ) -> None:
        self._queued_placeholder_map[(chat_id, message_id)] = placeholder_id
        self._queued_placeholder_timestamps[(chat_id, message_id)] = time.monotonic()

    def _clear_queued_placeholder(self, chat_id: int, message_id: int) -> None:
        self._queued_placeholder_map.pop((chat_id, message_id), None)
        self._queued_placeholder_timestamps.pop((chat_id, message_id), None)

    def _wrap_topic_work(self, key: str, work: Any) -> Any:
        conversation_id = None
        try:
            chat_id, thread_id, _scope = parse_topic_key(key)
            conversation_id = topic_key(chat_id, thread_id)
        except Exception:
            conversation_id = None

        if not conversation_id:
            return work

        async def wrapped() -> Any:
            token = set_conversation_id(conversation_id)
            try:
                return await work()
            finally:
                reset_conversation_id(token)

        return wrapped



from __future__ import annotations

import asyncio
import json
import logging
import time
from datetime import datetime
from pathlib import Path
from typing import Any, Awaitable, Callable, Optional

from ...core.logging_utils import log_event

TERMINAL_UPDATE_STATES = frozenset({"ok", "error", "rollback"})
RUNNING_UPDATE_STATES = frozenset({"running", "spawned"})

_NotifySender = Callable[[dict[str, Any], str], Awaitable[None]]
_ReadUpdateStatus = Callable[[], Optional[dict[str, Any]]]
_MarkUpdateNotified = Callable[[dict[str, Any]], None]
_FormatUpdateStatus = Callable[[Optional[dict[str, Any]]], str]
_SpawnTask = Callable[[Awaitable[None]], Any]


def _is_valid_notify_id(value: Any) -> bool:
    if isinstance(value, bool):
        return False
    if isinstance(value, int):
        return True
    return isinstance(value, str) and bool(value.strip())


def _coerce_notify_context(raw: Any) -> Optional[dict[str, Any]]:
    if not isinstance(raw, dict):
        return None
    chat_id = raw.get("chat_id")
    if not _is_valid_notify_id(chat_id):
        return None
    context: dict[str, Any] = {"chat_id": chat_id}
    thread_id = raw.get("thread_id")
    if _is_valid_notify_id(thread_id):
        context["thread_id"] = thread_id
    reply_to = raw.get("reply_to")
    if _is_valid_notify_id(reply_to):
        context["reply_to"] = reply_to
    return context


def _normalize_platform(platform: str) -> str:
    return str(platform or "").strip().lower()


def build_update_notify_metadata(
    *,
    platform: str,
    chat_id: int | str,
    thread_id: Optional[int | str] = None,
    reply_to: Optional[int | str] = None,
    include_legacy_telegram_keys: bool = False,
) -> dict[str, Any]:
    platform_key = _normalize_platform(platform)
    context: dict[str, Any] = {"chat_id": chat_id}
    if thread_id is not None:
        context["thread_id"] = thread_id
    if reply_to is not None:
        context["reply_to"] = reply_to
    coerced = _coerce_notify_context(context)
    if not platform_key or coerced is None:
        raise ValueError("Invalid update notification metadata.")

    payload: dict[str, Any] = {
        "notify_platform": platform_key,
        "notify_context": coerced,
    }
    if include_legacy_telegram_keys and platform_key == "telegram":
        chat_raw = coerced.get("chat_id")
        if isinstance(chat_raw, int) and not isinstance(chat_raw, bool):
            payload["notify_chat_id"] = chat_raw
        thread_raw = coerced.get("thread_id")
        if isinstance(thread_raw, int) and not isinstance(thread_raw, bool):
            payload["notify_thread_id"] = thread_raw
        reply_raw = coerced.get("reply_to")
        if isinstance(reply_raw, int) and not isinstance(reply_raw, bool):
            payload["notify_reply_to"] = reply_raw
    return payload


def extract_update_notify_context(
    status: Optional[dict[str, Any]],
    *,
    platform: str,
) -> Optional[dict[str, Any]]:
    if not isinstance(status, dict):
        return None
    platform_key = _normalize_platform(platform)
    notify_platform = status.get("notify_platform")
    notify_context = status.get("notify_context")
    if (
        isinstance(notify_platform, str)
        and _normalize_platform(notify_platform) == platform_key
        and isinstance(notify_context, dict)
    ):
        return _coerce_notify_context(notify_context)

    if platform_key != "telegram":
        return None
    legacy_chat_id = status.get("notify_chat_id")
    if not isinstance(legacy_chat_id, int) or isinstance(legacy_chat_id, bool):
        return None
    context: dict[str, Any] = {"chat_id": legacy_chat_id}
    legacy_thread_id = status.get("notify_thread_id")
    if isinstance(legacy_thread_id, int) and not isinstance(legacy_thread_id, bool):
        context["thread_id"] = legacy_thread_id
    legacy_reply_to = status.get("notify_reply_to")
    if isinstance(legacy_reply_to, int) and not isinstance(legacy_reply_to, bool):
        context["reply_to"] = legacy_reply_to
    return context


def format_update_status_message(status: Optional[dict[str, Any]]) -> str:
    if not status:
        return "No update status recorded."
    state = str(status.get("status") or "unknown")
    message = str(status.get("message") or "")
    timestamp = status.get("at")
    rendered_time = ""
    if isinstance(timestamp, (int, float)):
        rendered_time = datetime.fromtimestamp(timestamp).isoformat(timespec="seconds")
    lines = [f"Update status: {state}"]
    if message:
        lines.append(f"Message: {message}")
    if rendered_time:
        lines.append(f"Last updated: {rendered_time}")
    return "\n".join(lines)


def mark_update_status_notified(
    *,
    path: Path,
    status: dict[str, Any],
    logger: logging.Logger,
    log_event_name: str,
) -> None:
    updated = dict(status)
    updated["notify_sent_at"] = time.time()
    try:
        path.write_text(json.dumps(updated), encoding="utf-8")
    except Exception as exc:
        log_event(
            logger,
            logging.WARNING,
            log_event_name,
            exc=exc,
        )


class ChatUpdateStatusNotifier:
    def __init__(
        self,
        *,
        platform: str,
        logger: logging.Logger,
        read_status: _ReadUpdateStatus,
        send_notice: _NotifySender,
        spawn_task: _SpawnTask,
        mark_notified: _MarkUpdateNotified,
        format_status: Optional[_FormatUpdateStatus] = None,
        running_message: str = "Update still running.",
    ) -> None:
        self._platform = _normalize_platform(platform)
        self._logger = logger
        self._read_status = read_status
        self._send_notice = send_notice
        self._spawn_task = spawn_task
        self._mark_notified = mark_notified
        self._format_status = format_status or format_update_status_message
        self._running_message = running_message

    def build_spawn_metadata(
        self,
        *,
        chat_id: int | str,
        thread_id: Optional[int | str] = None,
        reply_to: Optional[int | str] = None,
        include_legacy_telegram_keys: bool = False,
    ) -> dict[str, Any]:
        return build_update_notify_metadata(
            platform=self._platform,
            chat_id=chat_id,
            thread_id=thread_id,
            reply_to=reply_to,
            include_legacy_telegram_keys=include_legacy_telegram_keys,
        )

    def schedule_watch(
        self,
        notify_context: dict[str, Any],
        *,
        timeout_seconds: float = 300.0,
        interval_seconds: float = 2.0,
    ) -> None:
        context = _coerce_notify_context(notify_context)
        if context is None:
            return

        async def _watch() -> None:
            deadline = time.monotonic() + timeout_seconds
            while time.monotonic() < deadline:
                status = self._read_status()
                if isinstance(status, dict):
                    state = str(status.get("status") or "")
                    if state in TERMINAL_UPDATE_STATES:
                        if status.get("notify_sent_at"):
                            return
                        resolved = (
                            extract_update_notify_context(
                                status, platform=self._platform
                            )
                            or context
                        )
                        await self._send_notice(
                            resolved,
                            self._format_status(status),
                        )
                        self._mark_notified(status)
                        return
                await asyncio.sleep(interval_seconds)
            await self._send_notice(context, self._running_message)

        self._spawn_task(_watch())

    async def maybe_send_notice(self) -> None:
        status = self._read_status()
        if not isinstance(status, dict):
            return
        context = extract_update_notify_context(status, platform=self._platform)
        if context is None:
            return
        if status.get("notify_sent_at"):
            return
        state = str(status.get("status") or "")
        if state in RUNNING_UPDATE_STATES:
            self.schedule_watch(context)
            return
        if state not in TERMINAL_UPDATE_STATES:
            return
        await self._send_notice(context, self._format_status(status))
        self._mark_notified(status)



from codex_autorunner.integrations.telegram.helpers import (
    _coerce_model_options,
    _format_model_list,
)


def test_coerce_model_options_uses_model_id_for_alias_display_name() -> None:
    options = _coerce_model_options(
        {
            "data": [
                {
                    "id": "gpt-5.3-codex-spark",
                    "displayName": "GPT-5.3-Codex-Spark",
                    "defaultReasoningEffort": "medium",
                    "supportedReasoningEfforts": ["low", "medium", "high"],
                }
            ]
        }
    )

    assert len(options) == 1
    assert options[0].label == "gpt-5.3-codex-spark (default medium)"


def test_format_model_list_uses_model_id_for_alias_display_name() -> None:
    result = _format_model_list(
        {
            "data": [
                {
                    "id": "gpt-5.3-codex-spark",
                    "displayName": "GPT-5.3-Codex-Spark",
                    "defaultReasoningEffort": "medium",
                    "supportedReasoningEfforts": ["low", "medium", "high"],
                }
            ]
        }
    )

    assert "gpt-5.3-codex-spark (GPT-5.3-Codex-Spark)" not in result
    assert "gpt-5.3-codex-spark [effort: low, medium, high] (default medium)" in result


def test_coerce_model_options_keeps_distinct_display_name() -> None:
    options = _coerce_model_options(
        {
            "data": [
                {
                    "id": "internal-preview-model",
                    "displayName": "Internal Preview (Fast)",
                    "defaultReasoningEffort": "medium",
                }
            ]
        }
    )

    assert len(options) == 1
    assert (
        options[0].label
        == "internal-preview-model (Internal Preview (Fast)) (default medium)"
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, Optional

TurnTriggerMode = Literal["always", "mentions"]


@dataclass(frozen=True)
class PlainTextTurnContext:
    """Shared trigger context for plain-text direct-chat turns."""

    text: str
    chat_type: Optional[str] = None
    bot_username: Optional[str] = None
    reply_to_is_bot: bool = False
    reply_to_username: Optional[str] = None
    reply_to_message_id: Optional[int] = None
    thread_id: Optional[int] = None


def should_trigger_plain_text_turn(
    *,
    mode: TurnTriggerMode,
    context: PlainTextTurnContext,
) -> bool:
    """Return True when a plain-text message should trigger an agent turn."""

    if mode == "always":
        return True
    if mode != "mentions":
        return False

    if context.chat_type == "private":
        return True

    lowered = (context.text or "").lower()
    if context.bot_username:
        needle = f"@{context.bot_username}".lower()
        if needle in lowered:
            return True

    implicit_topic_reply = (
        context.thread_id is not None
        and context.reply_to_message_id is not None
        and context.reply_to_message_id == context.thread_id
    )
    if context.reply_to_is_bot and not implicit_topic_reply:
        return True
    if (
        context.bot_username
        and context.reply_to_username
        and context.reply_to_username.lower() == context.bot_username.lower()
        and not implicit_topic_reply
    ):
        return True

    return False



from .archive_helpers import archive_flow_run_artifacts
from .controller import FlowController
from .definition import FlowDefinition, StepFn, StepOutcome
from .models import (
    FlowArtifact,
    FlowEvent,
    FlowEventType,
    FlowRunRecord,
    FlowRunStatus,
)
from .pause_dispatch import (
    PauseDispatchSnapshot,
    format_pause_reason,
    latest_dispatch_seq,
    load_latest_paused_ticket_flow_dispatch,
)
from .runtime import FlowRuntime
from .store import FlowStore

__all__ = [
    "FlowController",
    "FlowDefinition",
    "StepFn",
    "StepOutcome",
    "FlowArtifact",
    "FlowEvent",
    "FlowEventType",
    "FlowRunRecord",
    "FlowRunStatus",
    "PauseDispatchSnapshot",
    "FlowRuntime",
    "FlowStore",
    "archive_flow_run_artifacts",
    "format_pause_reason",
    "latest_dispatch_seq",
    "load_latest_paused_ticket_flow_dispatch",
]



from codex_autorunner.agents.opencode.logging import OpenCodeEventFormatter


def test_opencode_event_formatter_coalesces_reasoning_deltas() -> None:
    fmt = OpenCodeEventFormatter()
    part = {"id": "r1", "type": "reasoning"}

    # OpenCode can stream reasoning at token-level granularity. We should not emit
    # a log line per delta chunk.
    assert fmt.format_part("reasoning", part, "Hello ") == ["thinking"]
    assert fmt.format_part("reasoning", part, "world") == []

    # Flush once at end-of-turn.
    assert fmt.flush_all_reasoning() == ["**Hello world**"]
    assert fmt.flush_all_reasoning() == []


def test_opencode_event_formatter_emits_complete_lines_on_newlines() -> None:
    fmt = OpenCodeEventFormatter()
    part = {"id": "r1", "type": "reasoning"}

    # If the stream includes a newline, emit the completed line immediately,
    # leaving the remainder buffered for a final flush.
    assert fmt.format_part("reasoning", part, "Hello\nworld") == [
        "thinking",
        "**Hello**",
    ]
    assert fmt.flush_all_reasoning() == ["**world**"]



import asyncio  # noqa: F401
from pathlib import Path

import pytest

from codex_autorunner.agents.opencode.agent_config import (
    _build_agent_md,
    _extract_model_from_frontmatter,
    ensure_agent_config,
)
from codex_autorunner.agents.opencode.supervisor import OpenCodeSupervisor


def test_build_agent_md():
    """Test agent markdown generation with frontmatter."""
    result = _build_agent_md(
        agent_id="subagent",
        model="zai-coding-plan/glm-4.7-flashx",
        title="Subagent",
        description="Subagent for subagent tasks",
    )

    assert result.startswith("---")
    assert "agent: subagent" in result
    assert 'title: "Subagent"' in result
    assert 'description: "Subagent for subagent tasks"' in result
    assert "model: zai-coding-plan/glm-4.7-flashx" in result
    assert result.endswith("---\n")


def test_extract_model_from_frontmatter():
    """Test extracting model from YAML frontmatter."""
    content = """---
agent: subagent
title: "Subagent"
description: "Subagent for subagent tasks"
model: zai-coding-plan/glm-4.7-flashx
---
Some content here
"""
    result = _extract_model_from_frontmatter(content)
    assert result == "zai-coding-plan/glm-4.7-flashx"


def test_extract_model_from_frontmatter_no_model():
    """Test extracting model from frontmatter without model field."""
    content = """---
agent: subagent
title: "Subagent"
---
Some content here
"""
    result = _extract_model_from_frontmatter(content)
    assert result is None


def test_extract_model_from_frontmatter_invalid():
    """Test extracting model from invalid frontmatter."""
    content = """No frontmatter here
agent: subagent
---
Some content
"""
    result = _extract_model_from_frontmatter(content)
    assert result is None


@pytest.mark.anyio
async def test_ensure_agent_config_creates_file(tmp_path: Path):
    """Test that ensure_agent_config creates agent config file."""
    agent_dir = tmp_path / ".opencode" / "agent"
    agent_dir.mkdir(parents=True, exist_ok=True)

    await ensure_agent_config(
        workspace_root=tmp_path,
        agent_id="subagent",
        model="zai-coding-plan/glm-4.7-flashx",
        title="Subagent",
        description="Subagent for subagent tasks",
    )

    agent_file = agent_dir / "subagent.md"
    assert agent_file.exists()

    content = agent_file.read_text(encoding="utf-8")
    assert "agent: subagent" in content
    assert "model: zai-coding-plan/glm-4.7-flashx" in content


@pytest.mark.anyio
async def test_ensure_agent_config_skips_if_no_model(tmp_path: Path):
    """Test that ensure_agent_config skips if model is None."""
    agent_dir = tmp_path / ".opencode" / "agent"
    agent_dir.mkdir(parents=True, exist_ok=True)

    await ensure_agent_config(
        workspace_root=tmp_path,
        agent_id="subagent",
        model=None,
    )

    agent_file = agent_dir / "subagent.md"
    assert not agent_file.exists()


@pytest.mark.anyio
async def test_ensure_agent_config_skips_if_model_unchanged(tmp_path: Path):
    """Test that ensure_agent_config skips if model hasn't changed."""
    agent_dir = tmp_path / ".opencode" / "agent"
    agent_dir.mkdir(parents=True, exist_ok=True)
    agent_file = agent_dir / "subagent.md"

    existing_content = """---
agent: subagent
title: "Subagent"
description: "Subagent for subagent tasks"
model: zai-coding-plan/glm-4.7-flashx
---
"""
    agent_file.write_text(existing_content, encoding="utf-8")

    await ensure_agent_config(
        workspace_root=tmp_path,
        agent_id="subagent",
        model="zai-coding-plan/glm-4.7-flashx",
        title="Subagent",
        description="Subagent for subagent tasks",
    )

    content = agent_file.read_text(encoding="utf-8")
    assert content == existing_content


@pytest.mark.anyio
async def test_ensure_agent_config_updates_if_model_changed(tmp_path: Path):
    """Test that ensure_agent_config updates if model has changed."""
    agent_dir = tmp_path / ".opencode" / "agent"
    agent_dir.mkdir(parents=True, exist_ok=True)
    agent_file = agent_dir / "subagent.md"

    existing_content = """---
agent: subagent
title: "Subagent"
description: "Subagent for subagent tasks"
model: zai-coding-plan/glm-4.7
---
"""
    agent_file.write_text(existing_content, encoding="utf-8")

    await ensure_agent_config(
        workspace_root=tmp_path,
        agent_id="subagent",
        model="zai-coding-plan/glm-4.7-flashx",
        title="Subagent",
        description="Subagent for subagent tasks",
    )

    content = agent_file.read_text(encoding="utf-8")
    assert "model: zai-coding-plan/glm-4.7-flashx" in content
    assert content.count("model: zai-coding-plan/glm-4.7-flashx") == 1


def test_supervisor_ensure_subagent_config():
    """Test OpenCodeSupervisor.ensure_subagent_config calls agent_config."""
    supervisor = OpenCodeSupervisor(
        command=["opencode", "serve"],
        subagent_models={
            "subagent": "zai-coding-plan/glm-4.7-flashx",
            "other": "zai-coding-plan/glm-4.7",
        },
    )

    assert supervisor._subagent_models == {
        "subagent": "zai-coding-plan/glm-4.7-flashx",
        "other": "zai-coding-plan/glm-4.7",
    }



import asyncio
import logging
from pathlib import Path
from typing import Optional

from fastapi import APIRouter, HTTPException, Request
from fastapi.responses import JSONResponse

from ....core import update as update_core
from ....core.config import HubConfig
from ....core.self_describe import collect_describe_data
from ....core.update import (
    UpdateInProgressError,
    _normalize_update_ref,
    _normalize_update_target,
    _read_update_status,
    _spawn_update_process,
    _system_update_check,
)
from ....core.update_paths import resolve_update_paths
from ....core.utils import find_repo_root
from ..schemas import (
    SystemHealthResponse,
    SystemUpdateCheckResponse,
    SystemUpdateRequest,
    SystemUpdateResponse,
    SystemUpdateStatusResponse,
    SystemUpdateTargetOption,
    SystemUpdateTargetsResponse,
)
from ..static_assets import missing_static_assets
from ..static_refresh import refresh_static_assets

_pid_is_running = update_core._pid_is_running
_system_update_worker = update_core._system_update_worker
_update_lock_active = update_core._update_lock_active
_update_lock_path = update_core._update_lock_path
_update_status_path = update_core._update_status_path
_normalize_update_backend = update_core._normalize_update_backend
_resolve_update_backend = update_core._resolve_update_backend
_required_update_commands = update_core._required_update_commands
_available_update_target_options = update_core._available_update_target_options
_default_update_target = update_core._default_update_target
shutil = update_core.shutil
subprocess = update_core.subprocess
sys = update_core.sys


def build_system_routes() -> APIRouter:
    router = APIRouter()

    @router.get("/health", response_model=SystemHealthResponse)
    async def system_health(request: Request):
        try:
            config = request.app.state.config
        except AttributeError:
            config = None
        mode = "hub" if isinstance(config, HubConfig) else "repo"
        base_path = getattr(request.app.state, "base_path", "")
        asset_version = getattr(request.app.state, "asset_version", None)
        static_dir = getattr(getattr(request.app, "state", None), "static_dir", None)
        if not isinstance(static_dir, Path):
            return JSONResponse(
                {
                    "status": "error",
                    "detail": "Static UI assets missing; reinstall package",
                    "mode": mode,
                    "base_path": base_path,
                },
                status_code=500,
            )
        missing = await asyncio.to_thread(missing_static_assets, static_dir)
        if missing:
            if refresh_static_assets(request.app):
                static_dir = getattr(
                    getattr(request.app, "state", None), "static_dir", None
                )
                if isinstance(static_dir, Path):
                    missing = await asyncio.to_thread(missing_static_assets, static_dir)
                else:
                    missing = ["index.html"]
            if not missing:
                return {
                    "status": "ok",
                    "mode": mode,
                    "base_path": base_path,
                    "asset_version": asset_version,
                }
            return JSONResponse(
                {
                    "status": "error",
                    "detail": "Static UI assets missing; reinstall package",
                    "missing": missing,
                    "mode": mode,
                    "base_path": base_path,
                },
                status_code=500,
            )
        return {
            "status": "ok",
            "mode": mode,
            "base_path": base_path,
            "asset_version": asset_version,
        }

    @router.get("/system/update/check", response_model=SystemUpdateCheckResponse)
    async def system_update_check(request: Request):
        """
        Check if an update is available by comparing local git state vs remote.
        If local git state is unavailable, report that an update may be available.
        """
        try:
            config = request.app.state.config
        except AttributeError:
            config = None

        repo_url = "https://github.com/Git-on-my-level/codex-autorunner.git"
        repo_ref = "main"
        if config and isinstance(config, HubConfig):
            configured_url = getattr(config, "update_repo_url", None)
            if configured_url:
                repo_url = configured_url
            configured_ref = getattr(config, "update_repo_ref", None)
            if configured_ref:
                repo_ref = configured_ref

        try:
            return await asyncio.to_thread(
                _system_update_check, repo_url=repo_url, repo_ref=repo_ref
            )
        except Exception as e:
            logger = getattr(getattr(request.app, "state", None), "logger", None)
            if logger:
                logger.error("Update check error: %s", e, exc_info=True)
            raise HTTPException(status_code=500, detail=str(e)) from e

    @router.post("/system/update", response_model=SystemUpdateResponse)
    async def system_update(
        request: Request, payload: Optional[SystemUpdateRequest] = None
    ):
        """
        Pull latest code and refresh the running service.
        This will restart the server if successful.
        """
        try:
            config = request.app.state.config
        except AttributeError:
            config = None

        # Determine URL
        repo_url = "https://github.com/Git-on-my-level/codex-autorunner.git"
        repo_ref = "main"
        skip_checks = False
        update_backend = "auto"
        update_services: Optional[dict[str, str]] = None
        linux_hub_service_name = None
        linux_telegram_service_name = None
        linux_discord_service_name = None
        if config and isinstance(config, HubConfig):
            configured_url = getattr(config, "update_repo_url", None)
            if configured_url:
                repo_url = configured_url
            configured_ref = getattr(config, "update_repo_ref", None)
            if configured_ref:
                repo_ref = configured_ref
            skip_checks = bool(getattr(config, "update_skip_checks", False))
            update_backend = getattr(config, "update_backend", update_backend)
            update_services = getattr(config, "update_linux_service_names", None)
            if isinstance(update_services, dict):
                linux_hub_service_name = update_services.get("hub")
                linux_telegram_service_name = update_services.get("telegram")
                linux_discord_service_name = update_services.get("discord")
        elif config is not None:
            skip_checks = bool(getattr(config, "update_skip_checks", False))
            update_backend = getattr(config, "update_backend", update_backend)
            update_services = getattr(config, "update_linux_service_names", None)
            if isinstance(update_services, dict):
                linux_hub_service_name = update_services.get("hub")
                linux_telegram_service_name = update_services.get("telegram")
                linux_discord_service_name = update_services.get("discord")

        update_dir = resolve_update_paths(config=config).cache_dir

        try:
            target_raw = payload.target if payload else None
            if target_raw is None:
                target_raw = request.query_params.get("target")
            if target_raw is None:
                target_raw = _default_update_target(
                    raw_config=(config.raw if hasattr(config, "raw") else None),
                    update_backend=str(update_backend),
                    linux_service_names=(
                        update_services if isinstance(update_services, dict) else None
                    ),
                )
            update_target = _normalize_update_target(target_raw)
            logger = getattr(getattr(request.app, "state", None), "logger", None)
            if logger is None:
                logger = logging.getLogger("codex_autorunner.system_update")
            await asyncio.to_thread(
                _spawn_update_process,
                repo_url=repo_url,
                repo_ref=_normalize_update_ref(repo_ref),
                update_dir=update_dir,
                logger=logger,
                update_target=update_target,
                skip_checks=skip_checks,
                update_backend=update_backend,
                linux_hub_service_name=(
                    linux_hub_service_name
                    if isinstance(linux_hub_service_name, str)
                    else None
                ),
                linux_telegram_service_name=(
                    linux_telegram_service_name
                    if isinstance(linux_telegram_service_name, str)
                    else None
                ),
                linux_discord_service_name=(
                    linux_discord_service_name
                    if isinstance(linux_discord_service_name, str)
                    else None
                ),
            )
            return {
                "status": "ok",
                "message": f"Update started ({update_target}). Service will restart shortly.",
                "target": update_target,
            }
        except UpdateInProgressError as exc:
            raise HTTPException(status_code=409, detail=str(exc)) from exc
        except ValueError as exc:
            raise HTTPException(status_code=400, detail=str(exc)) from exc
        except Exception as e:
            logger = getattr(getattr(request.app, "state", None), "logger", None)
            if logger:
                logger.error("Update error: %s", e, exc_info=True)
            raise HTTPException(status_code=500, detail=str(e)) from e

    @router.get("/system/update/targets", response_model=SystemUpdateTargetsResponse)
    async def system_update_targets(request: Request):
        try:
            config = request.app.state.config
        except AttributeError:
            config = None

        update_backend = "auto"
        update_services: Optional[dict[str, str]] = None
        raw_config = getattr(config, "raw", None)
        if config is not None:
            update_backend = str(getattr(config, "update_backend", update_backend))
            raw_services = getattr(config, "update_linux_service_names", None)
            if isinstance(raw_services, dict):
                update_services = raw_services

        options = _available_update_target_options(
            raw_config=raw_config if isinstance(raw_config, dict) else None,
            update_backend=update_backend,
            linux_service_names=update_services,
        )
        default_target = _default_update_target(
            raw_config=raw_config if isinstance(raw_config, dict) else None,
            update_backend=update_backend,
            linux_service_names=update_services,
        )
        return {
            "targets": [
                SystemUpdateTargetOption(value=value, label=label)
                for value, label in options
            ],
            "default_target": default_target,
        }

    @router.get("/system/update/status", response_model=SystemUpdateStatusResponse)
    async def system_update_status():
        status = await asyncio.to_thread(_read_update_status)
        if status is None:
            return {"status": "unknown", "message": "No update status recorded."}
        return status

    @router.get("/system/describe")
    async def system_describe(request: Request):
        try:
            repo_root = find_repo_root(request.app.state.engine.repo_root)
            return await asyncio.to_thread(collect_describe_data, repo_root)
        except Exception as exc:
            logger = getattr(getattr(request.app, "state", None), "logger", None)
            if logger:
                logger.error("Describe endpoint error: %s", exc, exc_info=True)
            raise HTTPException(status_code=500, detail=str(exc)) from exc

    return router



import logging
from types import SimpleNamespace
from typing import Any

import pytest

from codex_autorunner.integrations.telegram.adapter import (
    TelegramCommand,
    TelegramMessage,
)
from codex_autorunner.integrations.telegram.handlers.commands_runtime import (
    TelegramCommandHandlers,
)
from tests.fixtures.telegram_command_helpers import make_command_spec

# Helper usage: use `make_command_spec(...)` for concise alias command setup while
# keeping alias-routing assertions explicit in each test.


class _AliasHarness(TelegramCommandHandlers):
    def __init__(self) -> None:
        self._logger = logging.getLogger(__name__)
        self._resume_options: dict[str, Any] = {}
        self._bind_options: dict[str, Any] = {}
        self._agent_options: dict[str, Any] = {}
        self._model_options: dict[str, Any] = {}
        self._model_pending: dict[str, Any] = {}
        self.sent_messages: list[str] = []
        self.model_calls: list[str] = []

        async def _handle_model_alias(
            _message: TelegramMessage, args: str, _runtime: Any
        ) -> None:
            self.model_calls.append(args)

        self._command_specs = {
            "model": make_command_spec(
                "model",
                "list or set the model",
                handler=_handle_model_alias,
            )
        }

    async def _resolve_topic_key(self, _chat_id: int, _thread_id: Any) -> str:
        return "1:root"

    async def _send_message(
        self,
        _chat_id: int,
        text: str,
        *,
        thread_id: Any = None,
        reply_to: Any = None,
    ) -> None:
        self.sent_messages.append(text)


def _message() -> TelegramMessage:
    return TelegramMessage(
        update_id=1,
        message_id=2,
        chat_id=3,
        thread_id=4,
        from_user_id=5,
        text="/models",
        date=0,
        is_topic_message=False,
    )


@pytest.mark.asyncio
async def test_models_alias_routes_to_model_handler() -> None:
    harness = _AliasHarness()
    runtime = SimpleNamespace(current_turn_id=None)
    command = TelegramCommand(name="models", args="list", raw="/models list")

    await harness._handle_command(command, _message(), runtime)

    assert harness.model_calls == ["list"]
    assert harness.sent_messages == []



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Optional, Protocol

from ..core.flows.models import FlowEventType

EmitEventFn = Callable[[FlowEventType, dict[str, Any]], None]


@dataclass(frozen=True)
class AgentTurnRequest:
    agent_id: str  # "codex" | "opencode"
    prompt: str
    workspace_root: Path
    conversation_id: Optional[str] = None
    # Optional, agent-specific extras.
    options: Optional[dict[str, Any]] = None
    # Optional flow event emitter (for live streaming).
    emit_event: Optional[EmitEventFn] = None
    # Optional list of additional messages to send in the same turn.
    # Each message is a dict with a "text" field. Agents that support
    # multiple messages will receive all of them; others may queue them.
    additional_messages: Optional[list[dict[str, Any]]] = None


@dataclass(frozen=True)
class AgentTurnResult:
    agent_id: str
    conversation_id: str
    turn_id: str
    text: str
    error: Optional[str] = None
    raw: Optional[dict[str, Any]] = None


class AgentPool(Protocol):
    """Port for ticket-flow agent execution."""

    async def run_turn(self, req: AgentTurnRequest) -> AgentTurnResult: ...

    async def close_all(self) -> None: ...



from __future__ import annotations

import json
import logging
from pathlib import Path
from typing import Any, Optional

from .locks import file_lock
from .time_utils import now_iso
from .utils import atomic_write

PMA_ACTIVE_SINK_FILENAME = "active_sink.json"

logger = logging.getLogger(__name__)


class PmaActiveSinkStore:
    def __init__(self, hub_root: Path) -> None:
        self._path = hub_root / ".codex-autorunner" / "pma" / PMA_ACTIVE_SINK_FILENAME

    def _lock_path(self) -> Path:
        return self._path.with_suffix(self._path.suffix + ".lock")

    def load(self) -> Optional[dict[str, Any]]:
        with file_lock(self._lock_path()):
            return self._load_unlocked()

    def set_web(self) -> dict[str, Any]:
        payload = {
            "version": 1,
            "kind": "web",
            "updated_at": now_iso(),
            "last_delivery_turn_id": None,
        }
        with file_lock(self._lock_path()):
            self._save_unlocked(payload)
        return payload

    def set_telegram(
        self,
        *,
        chat_id: int,
        thread_id: Optional[int],
        topic_key: Optional[str] = None,
    ) -> dict[str, Any]:
        target = (
            "telegram",
            str(int(chat_id)),
            str(int(thread_id)) if thread_id is not None else None,
        )
        payload: dict[str, Any] = {
            "version": 1,
            "kind": "telegram",
            "chat_id": int(chat_id),
            "thread_id": int(thread_id) if thread_id is not None else None,
            "updated_at": now_iso(),
            "last_delivery_turn_id": None,
        }
        if topic_key:
            payload["topic_key"] = topic_key
        with file_lock(self._lock_path()):
            existing = self._load_unlocked()
            payload["last_delivery_turn_id"] = self._last_delivery_for_target(
                existing, target
            )
            self._save_unlocked(payload)
        return payload

    def set_chat(
        self,
        platform: str,
        *,
        chat_id: str,
        thread_id: Optional[str] = None,
        conversation_key: Optional[str] = None,
    ) -> dict[str, Any]:
        platform_norm = str(platform or "").strip().lower()
        chat_id_norm = str(chat_id or "").strip()
        if not platform_norm:
            raise ValueError("platform must be non-empty")
        if not chat_id_norm:
            raise ValueError("chat_id must be non-empty")
        thread_id_norm = (
            str(thread_id).strip() if isinstance(thread_id, str) and thread_id else None
        )
        conversation_key_norm = (
            str(conversation_key).strip()
            if isinstance(conversation_key, str) and conversation_key
            else None
        )
        target = (platform_norm, chat_id_norm, thread_id_norm)
        with file_lock(self._lock_path()):
            existing = self._load_unlocked()
            last_delivery_turn_id = self._last_delivery_for_target(existing, target)
            payload: dict[str, Any] = {
                "version": 2,
                "kind": "chat",
                "platform": platform_norm,
                "chat_id": chat_id_norm,
                "thread_id": thread_id_norm,
                "updated_at": now_iso(),
                "last_delivery_turn_id": last_delivery_turn_id,
            }
            if conversation_key_norm:
                payload["conversation_key"] = conversation_key_norm
            self._save_unlocked(payload)
        return payload

    def clear(self) -> None:
        with file_lock(self._lock_path()):
            try:
                self._path.unlink()
            except FileNotFoundError:
                return
            except OSError as exc:
                logger.warning("Failed to clear PMA active sink: %s", exc)

    def mark_delivered(self, turn_id: str) -> bool:
        if not isinstance(turn_id, str) or not turn_id:
            return False
        with file_lock(self._lock_path()):
            payload = self._load_unlocked()
            if not isinstance(payload, dict):
                return False
            if payload.get("last_delivery_turn_id") == turn_id:
                return False
            payload["last_delivery_turn_id"] = turn_id
            payload["updated_at"] = now_iso()
            self._save_unlocked(payload)
        return True

    def _load_unlocked(self) -> Optional[dict[str, Any]]:
        if not self._path.exists():
            return None
        try:
            raw = self._path.read_text(encoding="utf-8")
        except OSError as exc:
            logger.warning("Failed to read PMA active sink: %s", exc)
            return None
        try:
            payload = json.loads(raw)
        except json.JSONDecodeError:
            return None
        if not isinstance(payload, dict):
            return None
        return payload

    def _save_unlocked(self, payload: dict[str, Any]) -> None:
        self._path.parent.mkdir(parents=True, exist_ok=True)
        atomic_write(self._path, json.dumps(payload, indent=2) + "\n")

    def _last_delivery_for_target(
        self,
        payload: Optional[dict[str, Any]],
        target: tuple[str, str, Optional[str]],
    ) -> Optional[str]:
        if not isinstance(payload, dict):
            return None
        payload_target = self._target_key(payload)
        if payload_target != target:
            return None
        last_delivery_turn_id = payload.get("last_delivery_turn_id")
        if isinstance(last_delivery_turn_id, str) and last_delivery_turn_id:
            return last_delivery_turn_id
        return None

    def _target_key(
        self, payload: dict[str, Any]
    ) -> Optional[tuple[str, str, Optional[str]]]:
        kind = payload.get("kind")
        if kind == "telegram":
            chat_id = payload.get("chat_id")
            if not isinstance(chat_id, int):
                return None
            thread_id = payload.get("thread_id")
            if thread_id is not None and not isinstance(thread_id, int):
                thread_norm: Optional[str] = None
            else:
                thread_norm = str(thread_id) if thread_id is not None else None
            return ("telegram", str(chat_id), thread_norm)

        if kind != "chat":
            return None
        platform = payload.get("platform")
        chat_id = payload.get("chat_id")
        if not isinstance(platform, str) or not platform.strip():
            return None
        if not isinstance(chat_id, str) or not chat_id.strip():
            return None
        thread_id = payload.get("thread_id")
        if isinstance(thread_id, str) and thread_id.strip():
            thread_norm = thread_id.strip()
        else:
            thread_norm = None
        return (platform.strip().lower(), chat_id.strip(), thread_norm)


__all__ = ["PmaActiveSinkStore", "PMA_ACTIVE_SINK_FILENAME"]



from codex_autorunner.integrations.telegram.helpers import _extract_thread_path


def test_extract_thread_path_prefers_nested_cwd_over_root() -> None:
    entry = {
        "root": "/hub",
        "workspace": {
            "cwd": "/repo",
        },
    }
    assert _extract_thread_path(entry) == "/repo"


def test_extract_thread_path_ignores_root_only() -> None:
    entry = {"root": "/hub"}
    assert _extract_thread_path(entry) is None



import logging
from pathlib import Path

from codex_autorunner.core.config import (
    CONFIG_FILENAME,
    REPO_OVERRIDE_FILENAME,
    load_repo_config,
)
from codex_autorunner.integrations.agents import opencode_supervisor_factory
from tests.conftest import write_test_config


def test_build_opencode_supervisor_from_repo_config(
    monkeypatch, tmp_path: Path
) -> None:
    hub_root = tmp_path / "hub"
    hub_root.mkdir()
    write_test_config(hub_root / CONFIG_FILENAME, {"mode": "hub"})

    repo_root = hub_root / "repo"
    repo_root.mkdir()
    write_test_config(
        repo_root / REPO_OVERRIDE_FILENAME,
        {
            "app_server": {
                "request_timeout": 91,
                "max_handles": 3,
                "idle_ttl_seconds": 120,
            },
            "opencode": {
                "server_scope": "global",
                "session_stall_timeout_seconds": 55,
                "max_text_chars": 9999,
                "max_handles": 7,
                "idle_ttl_seconds": 2222,
            },
            "agents": {
                "opencode": {
                    "binary": "/bin/opencode",
                    "serve_command": ["opencode", "serve", "--port", "0"],
                    "subagent_models": {"subagent": "model-x", "helper": "model-y"},
                }
            },
        },
    )

    config = load_repo_config(repo_root, hub_path=hub_root)
    captured: dict = {}

    def _fake_build_opencode_supervisor(**kwargs):
        captured.update(kwargs)
        return "supervisor"

    monkeypatch.setattr(
        opencode_supervisor_factory,
        "build_opencode_supervisor",
        _fake_build_opencode_supervisor,
    )

    logger = logging.getLogger("test.opencode")
    env = {"OPENCODE_SERVER_USERNAME": "alice"}
    supervisor = opencode_supervisor_factory.build_opencode_supervisor_from_repo_config(
        config,
        workspace_root=repo_root,
        logger=logger,
        base_env=env,
    )

    assert supervisor == "supervisor"
    assert captured["opencode_command"] == ["opencode", "serve", "--port", "0"]
    assert captured["opencode_binary"] == "/bin/opencode"
    assert captured["workspace_root"] == repo_root
    assert captured["logger"] is logger
    assert captured["request_timeout"] == 91
    assert captured["max_handles"] == 7
    assert captured["idle_ttl_seconds"] == 2222
    assert captured["server_scope"] == "global"
    assert captured["session_stall_timeout_seconds"] == 55
    assert captured["max_text_chars"] == 9999
    assert captured["base_env"] is env
    assert captured["subagent_models"] == {"subagent": "model-x", "helper": "model-y"}



from __future__ import annotations

import json
import logging
import time
from pathlib import Path

import pytest

import codex_autorunner.routes.system as system


@pytest.mark.parametrize(
    ("raw", "expected"),
    [
        (None, "both"),
        ("", "both"),
        ("ALL", "both"),
        ("web", "web"),
        ("ui", "web"),
        ("chat", "chat"),
        ("chat-apps", "chat"),
        ("telegram", "telegram"),
        ("tg", "telegram"),
        ("discord", "discord"),
        ("dc", "discord"),
    ],
)
def test_normalize_update_target(raw: str | None, expected: str) -> None:
    assert system._normalize_update_target(raw) == expected


def test_available_update_target_options_web_only_when_no_chat_available(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(
        system.update_core,
        "_chat_target_active",
        lambda **_kwargs: False,
    )
    options = system._available_update_target_options(
        raw_config={
            "telegram_bot": {"enabled": False},
            "discord_bot": {"enabled": False},
        },
        update_backend="systemd-user",
        linux_service_names={"hub": "car-hub"},
    )
    assert options == (("web", "Web only"),)
    assert (
        system._default_update_target(
            raw_config={
                "telegram_bot": {"enabled": False},
                "discord_bot": {"enabled": False},
            },
            update_backend="systemd-user",
            linux_service_names={"hub": "car-hub"},
        )
        == "web"
    )


def test_available_update_target_options_include_telegram_when_enableable(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(
        system.update_core,
        "_chat_target_active",
        lambda **_kwargs: False,
    )
    monkeypatch.setenv("CAR_TELEGRAM_BOT_TOKEN", "token")
    options = system._available_update_target_options(
        raw_config={
            "telegram_bot": {
                "enabled": True,
                "bot_token_env": "CAR_TELEGRAM_BOT_TOKEN",
            },
            "discord_bot": {"enabled": False},
        },
        update_backend="systemd-user",
        linux_service_names={"hub": "car-hub"},
    )
    assert options == (
        ("both", "Web + Chat Apps"),
        ("web", "Web only"),
        ("telegram", "Telegram only"),
    )


def test_available_update_target_options_include_discord_when_active(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(
        system.update_core,
        "_chat_target_active",
        lambda **kwargs: kwargs.get("target") == "discord",
    )
    options = system._available_update_target_options(
        raw_config={
            "telegram_bot": {"enabled": False},
            "discord_bot": {"enabled": False},
        },
        update_backend="systemd-user",
        linux_service_names={"hub": "car-hub", "discord": "car-discord"},
    )
    assert options == (
        ("both", "Web + Chat Apps"),
        ("web", "Web only"),
        ("discord", "Discord only"),
    )


@pytest.mark.parametrize(
    ("raw", "expected"),
    [
        (None, "auto"),
        ("", "auto"),
        ("AUTO", "auto"),
        ("launchd", "launchd"),
        ("systemd-user", "systemd-user"),
    ],
)
def test_normalize_update_backend(raw: str | None, expected: str) -> None:
    assert system._normalize_update_backend(raw) == expected


def test_resolve_update_backend_auto_linux(monkeypatch) -> None:
    monkeypatch.setattr(system.sys, "platform", "linux", raising=False)
    assert system._resolve_update_backend("auto") == "systemd-user"


@pytest.mark.parametrize(
    ("backend", "expected"),
    [
        ("launchd", ("git", "bash", "curl", "launchctl")),
        ("systemd-user", ("git", "bash", "curl", "systemctl")),
    ],
)
def test_required_update_commands(backend: str, expected: tuple[str, ...]) -> None:
    assert system._required_update_commands(backend) == expected


def test_update_lock_active_clears_stale(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.setenv("HOME", str(tmp_path))
    lock_path = system._update_lock_path()
    lock_path.parent.mkdir(parents=True, exist_ok=True)
    lock_path.write_text(json.dumps({"pid": 999999}), encoding="utf-8")

    monkeypatch.setattr(system, "_pid_is_running", lambda _pid: False)
    assert system._update_lock_active() is None
    assert not lock_path.exists()


def test_read_update_status_allows_recent_running_without_lock(
    tmp_path: Path, monkeypatch
) -> None:
    monkeypatch.setenv("HOME", str(tmp_path))
    status_path = system._update_status_path()
    status_path.parent.mkdir(parents=True, exist_ok=True)
    status_path.write_text(
        json.dumps(
            {
                "status": "running",
                "message": "Update spawned.",
                "at": time.time(),
            }
        ),
        encoding="utf-8",
    )

    monkeypatch.setattr(system.update_core, "_update_lock_active", lambda: None)

    payload = system._read_update_status()
    assert isinstance(payload, dict)
    assert payload["status"] == "running"


def test_read_update_status_marks_stale_running_without_lock(
    tmp_path: Path, monkeypatch
) -> None:
    monkeypatch.setenv("HOME", str(tmp_path))
    status_path = system._update_status_path()
    status_path.parent.mkdir(parents=True, exist_ok=True)
    status_path.write_text(
        json.dumps(
            {
                "status": "running",
                "message": "Update started.",
                "at": time.time() - 60,
            }
        ),
        encoding="utf-8",
    )

    monkeypatch.setattr(system.update_core, "_update_lock_active", lambda: None)

    payload = system._read_update_status()
    assert isinstance(payload, dict)
    assert payload["status"] == "error"
    assert payload["previous_status"] == "running"


def test_spawn_update_process_writes_status(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.setenv("HOME", str(tmp_path))
    calls: dict[str, object] = {}

    def fake_popen(cmd, cwd, start_new_session, stdout, stderr):  # type: ignore[no-untyped-def]
        calls["cmd"] = cmd
        calls["cwd"] = cwd
        return object()

    monkeypatch.setattr(system.subprocess, "Popen", fake_popen)

    update_dir = tmp_path / "update"
    logger = logging.getLogger("test")
    system._spawn_update_process(
        repo_url="https://example.com/repo.git",
        repo_ref="main",
        update_dir=update_dir,
        logger=logger,
        update_target="web",
        update_backend="systemd-user",
        linux_hub_service_name="car-hub",
        linux_telegram_service_name="car-telegram",
        linux_discord_service_name="car-discord",
        notify_platform="discord",
        notify_context={"chat_id": "channel-1"},
    )

    status_path = system._update_status_path()
    payload = json.loads(status_path.read_text(encoding="utf-8"))
    assert payload["status"] == "running"
    assert payload["notify_platform"] == "discord"
    assert payload["notify_context"] == {"chat_id": "channel-1"}
    assert "log_path" in payload
    cmd = calls["cmd"]
    assert "--repo-url" in cmd
    assert str(update_dir) in cmd
    assert "--backend" in cmd
    assert "systemd-user" in cmd
    assert "--hub-service-name" in cmd
    assert "car-hub" in cmd
    assert "--discord-service-name" in cmd
    assert "car-discord" in cmd


def test_system_update_worker_rejects_invalid_target(
    tmp_path: Path, monkeypatch
) -> None:
    monkeypatch.setenv("HOME", str(tmp_path))
    logger = logging.getLogger("test")
    update_dir = tmp_path / "update"

    system._system_update_worker(
        repo_url="https://example.com/repo.git",
        repo_ref="main",
        update_dir=update_dir,
        logger=logger,
        update_target="nope",
    )

    payload = json.loads(system._update_status_path().read_text(encoding="utf-8"))
    assert payload["status"] == "error"


def test_system_update_worker_rejects_invalid_backend(
    tmp_path: Path, monkeypatch
) -> None:
    monkeypatch.setenv("HOME", str(tmp_path))
    logger = logging.getLogger("test")
    update_dir = tmp_path / "update"

    system._system_update_worker(
        repo_url="https://example.com/repo.git",
        repo_ref="main",
        update_dir=update_dir,
        logger=logger,
        update_target="web",
        update_backend="invalid-backend",
    )

    payload = json.loads(system._update_status_path().read_text(encoding="utf-8"))
    assert payload["status"] == "error"
    assert "Unsupported update backend" in str(payload["message"])


def test_system_update_worker_missing_commands_releases_lock(
    tmp_path: Path, monkeypatch
) -> None:
    monkeypatch.setenv("HOME", str(tmp_path))
    monkeypatch.setattr(system.shutil, "which", lambda _cmd: None)
    logger = logging.getLogger("test")
    update_dir = tmp_path / "update"

    system._system_update_worker(
        repo_url="https://example.com/repo.git",
        repo_ref="main",
        update_dir=update_dir,
        logger=logger,
        update_target="web",
    )

    payload = json.loads(system._update_status_path().read_text(encoding="utf-8"))
    assert payload["status"] == "error"
    assert not system._update_lock_path().exists()


@pytest.mark.parametrize(
    ("backend", "missing_cmd"),
    [
        ("launchd", "launchctl"),
        ("systemd-user", "systemctl"),
    ],
)
def test_system_update_worker_backend_specific_missing_command(
    tmp_path: Path, monkeypatch, backend: str, missing_cmd: str
) -> None:
    monkeypatch.setenv("HOME", str(tmp_path))

    def fake_which(cmd: str) -> str | None:
        if cmd == missing_cmd:
            return None
        return f"/usr/bin/{cmd}"

    monkeypatch.setattr(system.shutil, "which", fake_which)
    logger = logging.getLogger("test")
    update_dir = tmp_path / "update"

    system._system_update_worker(
        repo_url="https://example.com/repo.git",
        repo_ref="main",
        update_dir=update_dir,
        logger=logger,
        update_target="web",
        update_backend=backend,
    )

    payload = json.loads(system._update_status_path().read_text(encoding="utf-8"))
    assert payload["status"] == "error"
    assert missing_cmd in str(payload["message"])


def test_system_update_worker_sets_helper_python_for_refresh_script(
    tmp_path: Path, monkeypatch
) -> None:
    monkeypatch.setenv("HOME", str(tmp_path))

    update_dir = tmp_path / "update"
    (update_dir / ".git").mkdir(parents=True)
    refresh_script = update_dir / "scripts" / "safe-refresh-local-linux-hub.sh"
    refresh_script.parent.mkdir(parents=True, exist_ok=True)
    refresh_script.write_text("#!/usr/bin/env bash\nexit 0\n", encoding="utf-8")

    monkeypatch.setattr(system.shutil, "which", lambda cmd: f"/usr/bin/{cmd}")
    monkeypatch.setattr(system.update_core, "_is_valid_git_repo", lambda _path: True)
    monkeypatch.setattr(system.update_core, "_run_cmd", lambda *_args, **_kwargs: None)

    captured_env: dict[str, str] = {}

    class _Proc:
        def __init__(self) -> None:
            self.stdout: list[str] = []
            self.returncode = 0

        def wait(self) -> int:
            return 0

    def fake_popen(cmd, cwd, env, stdout, stderr, text):  # type: ignore[no-untyped-def]
        captured_env.update(env)
        return _Proc()

    monkeypatch.setattr(system.subprocess, "Popen", fake_popen)
    monkeypatch.setattr(system.sys, "executable", "/opt/car/bin/python3", raising=False)
    logger = logging.getLogger("test")

    system._system_update_worker(
        repo_url="https://example.com/repo.git",
        repo_ref="main",
        update_dir=update_dir,
        logger=logger,
        update_target="web",
        update_backend="systemd-user",
        skip_checks=True,
        linux_discord_service_name="car-discord",
    )

    assert captured_env["HELPER_PYTHON"] == "/opt/car/bin/python3"
    assert captured_env["UPDATE_DISCORD_SERVICE_NAME"] == "car-discord"



from typing import TYPE_CHECKING, Optional

from fastapi import FastAPI, HTTPException

from ...core.usage import (
    UsageError,
    default_codex_home,
    get_hub_usage_series_cached,
    get_hub_usage_summary_cached,
    parse_iso_datetime,
)
from ...manifest import load_manifest
from .schemas import HubJobResponse

if TYPE_CHECKING:
    from .app_state import HubAppContext


def register_simple_hub_routes(app: FastAPI, context: "HubAppContext") -> None:
    @app.get("/hub/usage")
    def hub_usage(since: Optional[str] = None, until: Optional[str] = None):
        try:
            since_dt = parse_iso_datetime(since)
            until_dt = parse_iso_datetime(until)
        except UsageError as exc:
            raise HTTPException(status_code=400, detail=str(exc)) from exc

        manifest = load_manifest(context.config.manifest_path, context.config.root)
        repo_map = [
            (repo.id, (context.config.root / repo.path)) for repo in manifest.repos
        ]
        per_repo, unmatched, status = get_hub_usage_summary_cached(
            repo_map,
            default_codex_home(),
            config=context.config,
            since=since_dt,
            until=until_dt,
        )
        return {
            "mode": "hub",
            "hub_root": str(context.config.root),
            "codex_home": str(default_codex_home()),
            "since": since,
            "until": until,
            "status": status,
            "repos": [
                {
                    "id": repo_id,
                    "events": summary.events,
                    "totals": summary.totals.to_dict(),
                    "latest_rate_limits": summary.latest_rate_limits,
                }
                for repo_id, summary in per_repo.items()
            ],
            "unmatched": unmatched.to_dict(),
        }

    @app.get("/hub/usage/series")
    def hub_usage_series(
        since: Optional[str] = None,
        until: Optional[str] = None,
        bucket: str = "day",
        segment: str = "none",
    ):
        try:
            since_dt = parse_iso_datetime(since)
            until_dt = parse_iso_datetime(until)
        except UsageError as exc:
            raise HTTPException(status_code=400, detail=str(exc)) from exc

        manifest = load_manifest(context.config.manifest_path, context.config.root)
        repo_map = [
            (repo.id, (context.config.root / repo.path)) for repo in manifest.repos
        ]
        try:
            series, status = get_hub_usage_series_cached(
                repo_map,
                default_codex_home(),
                config=context.config,
                since=since_dt,
                until=until_dt,
                bucket=bucket,
                segment=segment,
            )
        except UsageError as exc:
            raise HTTPException(status_code=400, detail=str(exc)) from exc
        return {
            "mode": "hub",
            "hub_root": str(context.config.root),
            "codex_home": str(default_codex_home()),
            "since": since,
            "until": until,
            "status": status,
            **series,
        }

    @app.get("/hub/version")
    def hub_version():
        return {"asset_version": app.state.asset_version}

    @app.get("/hub/jobs/{job_id}", response_model=HubJobResponse)
    async def get_hub_job(job_id: str):
        job = await context.job_manager.get(job_id)
        if not job:
            raise HTTPException(status_code=404, detail="Job not found")
        return job.to_dict()



import logging
import shutil
from pathlib import Path

import pytest
from fastapi.testclient import TestClient

from codex_autorunner.core.config import load_hub_config
from codex_autorunner.server import create_hub_app
from codex_autorunner.surfaces.web import static_assets


def _write_required_assets(static_dir: Path) -> None:
    static_dir.mkdir(parents=True, exist_ok=True)
    (static_dir / "index.html").write_text(
        "<html>__CAR_ASSET_VERSION__</html>", encoding="utf-8"
    )
    (static_dir / "bootstrap.js").write_text(
        "console.log('bootstrap');", encoding="utf-8"
    )
    (static_dir / "loader.js").write_text("console.log('loader');", encoding="utf-8")
    (static_dir / "styles.css").write_text("body { }", encoding="utf-8")
    (static_dir / "app.js").write_text("console.log('app');", encoding="utf-8")
    (static_dir / "github.js").write_text("console.log('github');", encoding="utf-8")
    vendor_dir = static_dir / "vendor"
    vendor_dir.mkdir(parents=True, exist_ok=True)
    (vendor_dir / "xterm.js").write_text("console.log('xterm');", encoding="utf-8")
    (vendor_dir / "xterm-addon-fit.js").write_text(
        "console.log('fit');", encoding="utf-8"
    )
    (vendor_dir / "xterm.css").write_text("body { }", encoding="utf-8")


def test_materialize_static_assets_survives_source_removal(
    tmp_path: Path, monkeypatch
) -> None:
    source_dir = tmp_path / "source_static"
    _write_required_assets(source_dir)
    monkeypatch.setattr(static_assets, "resolve_static_dir", lambda: (source_dir, None))
    logger = logging.getLogger("test-static-assets")
    cache_root = tmp_path / "repo_root" / ".codex-autorunner" / "static-cache"
    cache_dir, cache_context = static_assets.materialize_static_assets(
        cache_root,
        max_cache_entries=5,
        max_cache_age_days=30,
        logger=logger,
    )
    assert cache_context is None
    assert cache_dir.exists()
    assert cache_dir.parent == cache_root
    shutil.rmtree(source_dir)
    assert static_assets.missing_static_assets(cache_dir) == []


def test_materialize_static_assets_falls_back_to_existing_cache(
    tmp_path: Path, monkeypatch
) -> None:
    cache_root = tmp_path / "repo_root" / ".codex-autorunner" / "static-cache"
    existing_cache = cache_root / "existing"
    _write_required_assets(existing_cache)
    source_dir = tmp_path / "missing_source"
    monkeypatch.setattr(static_assets, "resolve_static_dir", lambda: (source_dir, None))
    logger = logging.getLogger("test-static-assets")
    cache_dir, cache_context = static_assets.materialize_static_assets(
        cache_root,
        max_cache_entries=5,
        max_cache_age_days=30,
        logger=logger,
    )
    assert cache_context is None
    assert cache_dir == existing_cache


def test_materialize_static_assets_hard_fails_without_cache(
    tmp_path: Path, monkeypatch
) -> None:
    source_dir = tmp_path / "missing_source"
    monkeypatch.setattr(static_assets, "resolve_static_dir", lambda: (source_dir, None))
    logger = logging.getLogger("test-static-assets")
    cache_root = tmp_path / "repo_root" / ".codex-autorunner" / "static-cache"
    with pytest.raises(RuntimeError):
        static_assets.materialize_static_assets(
            cache_root,
            max_cache_entries=5,
            max_cache_age_days=30,
            logger=logger,
        )


def test_materialize_static_assets_prunes_old_entries(
    tmp_path: Path, monkeypatch
) -> None:
    source_dir = tmp_path / "source_static"
    _write_required_assets(source_dir)
    cache_root = tmp_path / "repo_root" / ".codex-autorunner" / "static-cache"
    old_cache = cache_root / "old"
    older_cache = cache_root / "older"
    _write_required_assets(old_cache)
    _write_required_assets(older_cache)
    monkeypatch.setattr(static_assets, "resolve_static_dir", lambda: (source_dir, None))
    logger = logging.getLogger("test-static-assets")
    cache_dir, cache_context = static_assets.materialize_static_assets(
        cache_root,
        max_cache_entries=1,
        max_cache_age_days=None,
        logger=logger,
    )
    assert cache_context is None
    entries = [
        path
        for path in cache_root.iterdir()
        if path.is_dir() and not path.name.startswith(".")
    ]
    assert cache_dir in entries
    assert len(entries) <= 1


def test_repo_app_serves_cached_static_assets(
    hub_env, tmp_path: Path, monkeypatch
) -> None:
    source_dir = tmp_path / "source_static"
    _write_required_assets(source_dir)
    monkeypatch.setattr(static_assets, "resolve_static_dir", lambda: (source_dir, None))
    app = create_hub_app(hub_env.hub_root)
    client = TestClient(app)
    shutil.rmtree(source_dir)
    res = client.get(f"/repos/{hub_env.repo_id}/")
    assert res.status_code == 200
    static_res = client.get(f"/repos/{hub_env.repo_id}/static/app.js")
    assert static_res.status_code == 200


def test_static_assets_cached_and_compressed(
    hub_env, tmp_path: Path, monkeypatch
) -> None:
    source_dir = tmp_path / "source_static"
    _write_required_assets(source_dir)
    (source_dir / "big.js").write_text("a" * 2048, encoding="utf-8")
    monkeypatch.setattr(static_assets, "resolve_static_dir", lambda: (source_dir, None))
    app = create_hub_app(hub_env.hub_root)
    client = TestClient(app)
    cache_res = client.get(f"/repos/{hub_env.repo_id}/static/app.js")
    assert cache_res.status_code == 200
    cache_control = cache_res.headers.get("Cache-Control", "")
    assert "max-age=31536000" in cache_control
    gzip_res = client.get(
        f"/repos/{hub_env.repo_id}/static/big.js", headers={"Accept-Encoding": "gzip"}
    )
    assert gzip_res.status_code == 200
    assert gzip_res.headers.get("Content-Encoding") == "gzip"


def test_repo_app_falls_back_to_hub_static_cache(
    hub_env, tmp_path: Path, monkeypatch
) -> None:
    hub_config = load_hub_config(hub_env.hub_root)
    hub_cache_root = hub_config.static_assets.cache_root
    existing_cache = hub_cache_root / "existing"
    _write_required_assets(existing_cache)
    source_dir = tmp_path / "missing_source"
    monkeypatch.setattr(static_assets, "resolve_static_dir", lambda: (source_dir, None))
    app = create_hub_app(hub_env.hub_root)
    client = TestClient(app)
    res = client.get(f"/repos/{hub_env.repo_id}/static/app.js")
    assert res.status_code == 200
